<?php
/**
 * $Horde: ansel/lib/Ansel.php,v 1.149 2007/02/06 18:01:18 chuck Exp $
 *
 * @package Ansel
 */

/** Horde_DataTree */
require_once 'Horde/DataTree.php';

/**
 * Ansel Base Class.
 *
 * @author  Chuck Hagenbuch <chuck@horde.org>
 * @package Ansel
 */
class Ansel {

    /**
     */
    function getDb()
    {
        require_once 'MDB2.php';
        $config = $GLOBALS['conf']['sql'];
        unset($config['charset']);
        return MDB2::factory($config);
    }

    /**
     * Retrieves the current user's gallery list from storage.
     *
     * @param integer $perm       The level of permissions to require for a
     *                            gallery to return it.
     * @param mixed $attributes   Restrict the galleries counted to those
     *                            matching $attributes. An array of
     *                            attribute/values pairs or a gallery owner
     *                            username.
     * @param string $parent      The parent share to start listing at.
     * @param boolean $allLevels  Return all levels, or just the direct
     *                            children of $parent? Defaults to all levels.
     * @param integer $from       The gallery to start listing at.
     * @param integer $count      The number of galleries to return.
     *
     * @return array  Returns a list of matching galleries.
     */
    function listGalleries($perm = PERMS_SHOW,
                           $attributes = null,
                           $parent = DATATREE_ROOT,
                           $allLevels = true,
                           $from = 0,
                           $count = 0)
    {
        global $ansel_shares, $notification, $prefs;

        /* If $parent is DATATREE_ROOT, then pass it through,
         * otherwise, treat it as a datatree id and retrieve the share
         * name before passing it on */
        if ($parent != DATATREE_ROOT) {
            $parent = $ansel_shares->getName($parent);
        }

        $galleries = $ansel_shares->listShares(Auth::getAuth(), $perm, $attributes, $parent, $allLevels, $from, $count);
        if (is_a($galleries, 'PEAR_Error')) {
            $notification->push(sprintf(_("An error occurred listing galleries: %s"), $galleries->getMessage()), 'horde.error');
            Horde::logMessage($galleries->getMessage(), __FILE__, __LINE__, PEAR_LOG_ERR);
            return array();
        }

        return $galleries;
    }

    /**
     * Returns a random gallery.
     *
     * @param integer $perm       The level of permissions to require for a
     *                            gallery to return it.
     * @param mixed $attributes   Restrict the galleries counted to those
     *                            matching $attributes. An array of
     *                            attribute/values pairs or a gallery owner
     *                            username.
     * @param string $parent      The parent share to start looking at.
     * @param boolean $allLevels  Return all levels, or just the direct
     *                            children of $parent? Defaults to all levels.
     *
     * @return object  Random gallery object
     */
    function &getRandomGallery($perm = PERMS_SHOW, $attributes = null, $parent = DATATREE_ROOT, $allLevels = true)
    {
        $num_galleries = Ansel::countGalleries($perm, $attributes, $parent, $allLevels);
        if (!$num_galleries) {
            return null;
        }

        $galleries = Ansel::listGalleries($perm, $attributes, $parent, $allLevels, rand(0, $num_galleries - 1), 1);
        list(,$gallery) = each($galleries);

        return $gallery;
    }

    /**
     * Counts the number of current user's gallery list from storage.
     *
     * @param integer $perm       The level of permissions to require for a
     *                            gallery to return it.
     * @param mixed $attributes   Restrict the galleries counted to those
     *                            matching $attributes. An array of
     *                            attribute/values pairs or a gallery owner
     *                            username.
     * @param string $parent      The parent share to start counting at.
     * @param boolean $allLevels  Return all levels, or just the direct
     *                            children of $parent? Defaults to all levels.
     *
     * @return integer  The number of matching galleries.
     */
    function countGalleries($perm = PERMS_SHOW,
                            $attributes = null,
                            $parent = DATATREE_ROOT,
                            $allLevels = true)
    {
        global $ansel_shares, $notification;

        $galleries = $ansel_shares->countShares(Auth::getAuth(), $perm, $attributes, $parent, $allLevels);
        if (is_a($galleries, 'PEAR_Error')) {
            $notification->push(sprintf(_("An error occurred counting galleries: %s"), $galleries->getMessage()), 'horde.error');
            return 0;
        }
        return $galleries;
    }

    /**
     * Return a string containing an <option> listing of the given
     * gallery array.
     *
     * @param array $selected     The gallery_id of the  gallery that is
     *                            selected by default in the returned option
     *                            list.
     * @param integer $perm       The permissions filter to use.
     * @param mixed $attributes   Restrict the galleries returned to those
     *                            matching $attributes. An array of
     *                            attribute/values pairs or a gallery owner
     *                            username.
     * @param string $parent      The parent share to start listing at.
     * @param integer $from       The gallery to start listing at.
     * @param integer $count      The number of galleries to return.
     *
     * @return string  The <option> list.
     */
    function selectGalleries($selected = null, $perm = PERMS_SHOW, $attributes = null, $parent = DATATREE_ROOT, $allLevels = true, $from = 0, $count = 0)
    {
        global $ansel_shares;
        require_once 'Horde/Tree.php';
        $galleries = Ansel::listGalleries($perm, $attributes, $parent, $allLevels, $from, $count);
        $tree = Horde_Tree::factory('gallery_tree', 'select');
        foreach ($galleries as $gallery_id => $gallery) {
            $indent = count($gallery->getParents());
            $label = $gallery->get('name');
            $params['selected'] = ($gallery_id == $selected) ? true : false;

            $parent = $ansel_shares->_datatree->getParent($gallery);
            if ($parent == DATATREE_ROOT) {
                $parent = null;
            }

            $tree->addNode($gallery->getId(), $parent, $label, $indent, true, $params);
        }

        return $tree->getTree();
    }

    /**
     * Create and initialise a new gallery object.
     *
     * @param string $gallery_name  The DataTree name of the new gallery.
     * @param array $attributes     The gallery attributes
     * @param object Perms $perm    The permissions for the gallery if the
     *                              defaults are not desirable.
     *
     * @return Ansel_Gallery  A new gallery object or PEAR_Error.
     */
    function &createGallery($gallery_name = '', $attributes = array(), $perm = null)
    {
        global $ansel_shares;
        if ($ansel_shares->exists($gallery_name)) {
            $error = sprintf(_("Gallery Unique name \"%s\" already exists."), $gallery_name);
            Horde::logMessage($error, __FILE__, __LINE__, PEAR_LOG_ERR);
            return PEAR::raiseError($error);
        }

        $gallery = &$ansel_shares->newGallery($gallery_name);
        if (is_a($gallery, 'PEAR_Error')) {
            return $gallery;
        }

        // Required values.
        if (empty($attributes['owner'])) {
            $attributes['owner'] = Auth::getAuth();
        }
        if (empty($attributes['name'])) {
            $attributes['name'] = _("Unnamed");
        }
        if (empty($attributes['desc'])) {
            $attributes['desc'] = '';
        }

        foreach ($attributes as $attribute => $value) {
            $gallery->set($attribute, $value);
        }

        $id = $ansel_shares->addGallery($gallery);
        if (is_a($id, 'PEAR_Error')) {
            $error = sprintf(_("The gallery \"%s\" could not be created: %s"), $attributes['name'], $id->getMessage());
            Horde::logMessage($error, __FILE__, __LINE__, PEAR_LOG_ERR);
            return PEAR::raiseError($error);
        }

        // Set initial permissions.
        if (empty($perm)) {
            $perm = $gallery->getPermission();
            $perm->addGuestPermission(PERMS_SHOW | PERMS_READ, false);
            $perm->addDefaultPermission(PERMS_SHOW | PERMS_READ, false);
        }

        $gallery = &$ansel_shares->getGallery($id);
        $gallery->setPermission($perm);
        $gallery->save();

        return $gallery;
    }

    /**
     * Return a link to an image placeholder, suitable for use in an <img/>
     * tag (or a Horde::img() call, with the path parameter set to * '').
     * This image should be used as a placeholder if the correct image can't
     * be retrieved
     *
     * @param string $view  The view ('screen', 'thumb', or 'full') to show.
     *                      Defaults to 'screen'.
     *
     * @return string  The image path.
     */
    function getErrorImage($view = 'screen')
    {
        return $GLOBALS['registry']->getImageDir() . '/' . $view . '-error.png';
    }

    /**
     * Return a link to an image, suitable for use in an <img/> tag
     * (or a Horde::img() call, with the path parameter set to
     * ''). Takes into account $conf['vfs']['direct'] and other
     * factors.
     *
     * @param string $imageId  The id of the image.
     * @param string $view     The view ('screen', 'thumb', or 'full') to show.
     * @param boolean $full    Return a path that includes the server name?
     *
     * @return string  The image path.
     */
    function getImageUrl($imageId, $view = 'screen', $full = false)
    {
        global $conf, $ansel_shares;

        if (empty($imageId) || $imageId == DATATREE_ROOT) {
            return Ansel::getErrorImage($view);
        } elseif (empty($conf['vfs']['direct'])) {
            $url = Util::addParameter('img/' . $view . '.php', 'image', $imageId);
            return Horde::applicationUrl($url, $full);
        }

        // We have to make sure the image exists first, since we won't
        // be going through img/*.php to auto-create it.
        if (is_a($image = $ansel_shares->getImage($imageId), 'PEAR_Error')) {
            return Ansel::getErrorImage($view);
        }
        if (is_a($result = $image->createView($view), 'PEAR_Error')) {
            return Ansel::getErrorImage($view);
        }

        $path = substr($imageId, -2) . '/' . $view;
        if ($full && substr($conf['vfs']['direct'], 0, 7) != 'http://') {
            return Horde::url($conf['vfs']['direct'] . $path . '/' . rawurlencode($image->getVFSName($view)), true, -1);
        } else {
            return $conf['vfs']['direct'] . $path . '/' . rawurlencode($image->getVFSName($view));
        }
    }

    /**
     */
    function getImageObject()
    {
        global $conf;
        require_once 'Horde/Image.php';
        $params = array('type' => $conf['image']['type'], 'temp' => Horde::getTempDir());
        $driver = empty($conf['image']['convert']) ? 'gd' : 'im';
        return Horde_Image::factory($driver, $params);
    }

    /**
     * Read an image from the filesystem.
     *
     * @param string $file     The filename of the image.
     * @param array $override  Overwrite the file array with these values.
     *
     * @return array  The image data of the file as an array or PEAR_Error
     */
    function &getImageFromFile($file, $override = array())
    {
        if (!file_exists($file)) {
            return PEAR::raiseError(sprintf(_("The file \"%s\" doesn't exist."), $file));
        }

        global $conf;

        // Get the mime type of the file (and make sure it's an image).
        require_once 'Horde/MIME/Magic.php';
        $mime_type = MIME_Magic::analyzeFile($file, isset($conf['mime']['magic_db']) ? $conf['mime']['magic_db'] : null);
        if (strpos($mime_type, 'image') === false) {
            return PEAR::raiseError(sprintf(_("Can't get unknown file type \"%s\"."), $file));
        }

        $image = array('image_filename' => basename($file),
                       'image_caption' => '',
                       'image_type' => $mime_type,
                       'data' => file_get_contents($file),
                       );

        // Override the array, for example if we're setting the filename to
        // something else.
        if (count($override)) {
            $image = array_merge($image, $override);
        }

        return $image;
    }

    /**
     * Check to see if a particular image manipulation function is
     * available.
     *
     * @param string $feature  The name of the function.
     *
     * @return boolean  True if the function is available.
     */
    function isAvailable($feature)
    {
        static $capabilities;

        if (!isset($capabilities)) {
            $im = Ansel::getImageObject();
            $capabilities = $im->getCapabilities();
        }

        return in_array($feature, $capabilities);
    }

    /**
     * Build Ansel's list of menu items.
     */
    function getMenu($returnType = 'object')
    {
        global $conf, $registry, $ansel_shares;

        require_once 'Horde/Menu.php';

        $menu = new Menu();
        $menu->add(Horde::applicationUrl('list.php'), _("_Galleries"), 'mypics.png', null, null, null, (basename($_SERVER['PHP_SELF']) == 'index.php' || basename($_SERVER['PHP_SELF']) == 'list.php' || basename($_SERVER['PHP_SELF']) == 'group.php') && !Util::getFormData('gallery') && Util::getFormData('owner') !== Auth::getAuth() ? 'current' : '__noselection');
        if (Auth::getAuth()) {
            $menu->add(Horde::applicationUrl(Util::addParameter('list.php', 'owner', Auth::getAuth())), _("_My Galleries"), 'mygalleries.png', null, null, null, (Util::getFormData('owner', false) == Auth::getAuth()) ? 'current' : '__noselection');
        }

        /* Let authenticated users create new galleries. */
        if (Auth::isAdmin() ||
            !$GLOBALS['perms']->exists('ansel') ||
            $GLOBALS['perms']->hasPermission('ansel', Auth::getAuth(), PERMS_EDIT)) {
            $menu->add(Horde::applicationUrl(Util::addParameter('gallery.php', 'actionID', 'add')), _("_New Gallery"), 'add.png', null, null, null, basename($_SERVER['PHP_SELF']) == 'gallery.php' ? (Util::getFormData('gallery') ? '__noselection' : 'current') : null);
        }

        /* Print. */
        if ($conf['menu']['print'] && ($pl = Util::nonInputVar('print_link'))) {
            $menu->add($pl, _("_Print"), 'print.png', $registry->getImageDir('horde'), '_blank', 'popup(this.href); return false;');
        }

        if ($returnType == 'object') {
            return $menu;
        } else {
            return $menu->render();
        }
    }

    /**
     * Generate a list of breadcrumbs showing where we are in the
     * gallery tree.
     */
    function getBreadCrumbs($separator = ' &raquo; ')
    {
        global $prefs, $ansel_shares;

        $groupby = Util::getFormData('groupby', $prefs->getValue('groupby'));
        $owner = Util::getFormData('owner');
        $gallery_id = Util::getFormData('gallery');
        $image_id = Util::getFormData('image');
        $actionID = Util::getFormData('actionID');
        $page = Util::getFormData('page', 0);

        if (!empty($gallery_id)) {
            $gallery = &$ansel_shares->getGallery($gallery_id);
            $owner = $gallery->get('owner');
        }

        if (!empty($image_id)) {
            $image = $ansel_shares->getImage($image_id);
        }

        if (!empty($owner)) {
            if ($owner == Auth::getAuth()) {
                $owner_title = _("My Galleries");
            } else {
                require_once('Horde/Identity.php');
                $identity = &Identity::singleton('none', $owner);
                $fullname = $identity->getValue('fullname');
                if (!$fullname) {
                    $fullname = $owner;
                }
                $owner_title = sprintf(_("%s's Galleries"), $fullname);
            }
        }

        $levels = 0;
        $nav = '<div class="breadcrumbs"><p>' . Horde::link(Horde::applicationUrl('list.php')) . _("Galleries") . '</a>';

        if (!empty($owner_title)) {
            $owner_title = htmlspecialchars($owner_title);
            $levels++;
            if (!empty($gallery_id)) {
                $nav .= $separator . Horde::link(Horde::applicationUrl(Util::addParameter('list.php', array('groupby' => $groupby, 'owner' => $owner)))) . $owner_title . '</a>';
            } else {
                $nav .= $separator . $owner_title;
            }
        }

        if (!empty($gallery_id)) {
            $levels++;

            $parent_list = $gallery->getParents();
            foreach ($parent_list as $id) {
                $p = $ansel_shares->getGallery($id);
                $nav .= $separator . Horde::link(Horde::applicationUrl(Util::addParameter('view.php', array('view' => 'Gallery', 'gallery' => $p->getId())))) . $p->get('name') . '</a>';
            }

            $nav .= $separator . Horde::link(Horde::applicationUrl(Util::addParameter('view.php', array('view' => 'Gallery', 'gallery' => $gallery_id, 'page' => $page)))) . htmlspecialchars($gallery->get('name')) . '</a>';
        }

        if (!empty($image_id)) {
            $levels++;
            $nav .= $separator . '<a id="breadcrumb_image" href="' . Horde::applicationUrl(Util::addParameter('view.php', array('gallery' => $gallery_id, 'image' => $image_id, 'view' => 'Image', 'page' => $page))) . '">' . htmlspecialchars($image->filename) . '</a>';
        }

        $nav .= '</p></div>';

        if ($levels) {
            return $nav;
        } else {
            return '';
        }
    }

}

/**
 * Copy and extension of the Horde_Share class for managing Ansel
 * Galleries as Horde shared objects. This class is independent from
 * Horde_Share for now so it can handle hierarchical shares which are
 * being phased out of the main Share code.
 *
 * @author  Chuck Hagenbuch <chuck@horde.org>
 * @author  Ben Chavet <ben@horde.org>
 * @since   Ansel 0.1
 * @package Ansel
 */
class Ansel_Shares {

    /**
     * Pointer to a DataTree instance to manage/store shares
     *
     * @var DataTree
     */
    var $_datatree;

    /**
     * The application we're managing shares for.
     *
     * @var string
     */
    var $_app = 'ansel';

    /**
     * The subclass of DataTreeObject to instantiate shares as.
     *
     * @var string
     */
    var $_shareObject = 'Ansel_Gallery';

    /**
     * A cache of all shares that have been retrieved, so we don't hit the
     * backend again and again for them.
     *
     * @var array
     */
    var $_cache = array();

    /**
     * Id-name-map of already cached share objects.
     *
     * @var array
     */
    var $_shareMap = array();

    /**
     * Cache used for listShares().
     *
     * @var array
     */
    var $_listcache = array();

    /**
     * Caches the number of share matching certain criteria.
     *
     * @see countShares()
     * @var array
     */
    var $_counts = array();

    /**
     * Cache of parent shares.
     *
     * This is an array with names as keys and the integer share id of the
     * direct parent as values.
     *
     * @see getShareParent()
     * @var array
     */
    var $_shareParents = array();

    /**
     * Cache if shares have children with share IDs as keys.
     *
     * @see hasChildren()
     * @var array
     */
    var $_hasChildren = array();

    /**
     * A list of objects that we're currently sorting, for reference during the
     * sorting algorithm.
     *
     * @var array
     */
    var $_sortList;

    /**
     * Constructor.
     */
    function Ansel_Shares()
    {
        global $conf, $registry;

        if (empty($conf['datatree']['driver'])) {
            Horde::fatal('You must configure a DataTree backend to use Shares.', __FILE__, __LINE__);
        }

        $driver = $conf['datatree']['driver'];
        $this->_datatree = &DataTree::singleton(
            $driver,
            array_merge(Horde::getDriverConfig('datatree', $driver), array('group' => 'horde.shares.' . $this->_app))
        );

        Horde::callHook('_horde_hook_share_init', array($this, $this->_app));
    }

    /**
     * Return a new gallery object.
     *
     * @param string $name  Gallery's DataTree name.
     *
     * @return Ansel_Gallery  A new gallery object.
     */
    function &newGallery($name)
    {
        if (empty($name)) {
            return PEAR::raiseError(_("Gallery names must be non-empty"));
        }
        $gallery = new Ansel_Gallery($name);
        $gallery->setShareOb($this);
        return $gallery;
    }

    /**
     * Return a Ansel_Gallery object corresponding to the
     * given name/ID.
     *
     * @param integer $id  The ID of the gallery to retrieve.
     *
     * @return Ansel_Gallery  The gallery object corresponding to the
     *                                 given name.
     */
    function &getGallery($id)
    {
        if (!isset($this->_cache[$id])) {
            $this->_cache[$id] = &$this->_datatree->getObjectById($id, 'Ansel_Gallery');
            if (!is_a($this->_cache[$id], 'PEAR_Error')) {
                $this->_cache[$id]->setShareOb($this);
            }
        }

        return $this->_cache[$id];
    }

    /**
     * Returns the image corresponding to the given id.
     *
     * @param integer $id  The ID of the image to retrieve.
     *
     * @return Ansel_Image  The image object corresponding to the given name.
     */
    function getImage($id)
    {
        $image = $GLOBALS['ansel_db']->queryRow('SELECT image_id, gallery_id, image_filename, image_type, image_caption, image_uploaded, image_sort FROM ansel_images WHERE image_id = ' . (int)$id, null, MDB2_FETCHMODE_ASSOC);
        if (is_null($image)) {
            return PEAR::raiseError('not found');
        } elseif (is_a($image, 'PEAR_Error')) {
            return $image;
        } else {
            return new Ansel_Image($image);
        }
    }

    /**
     * Returns the DataTree name for the given ID.
     *
     * @param integer $id  The ID of the DataTreeObject.
     *
     * @return string  The name of the DataTreeObject.
     */
    function getName($id)
    {
        return $this->_datatree->getName($id);
    }

    /**
     * Returns the datatree_id from the given DataTree name.
     *
     * @param string $name  The DataTreeObject name.
     *
     * @return mixed  The datatree_id | PEAR_Error
     */
     function getIdFromName($name)
     {
         return $this->_datatree->getId($name);
     }

    /**
     * Adds a new Ansel_Gallery to the object backend.
     *
     * @param Ansel_Gallery $gallery  The gallery to add.
     */
    function addGallery($gallery)
    {
        if (!is_a($gallery, 'Ansel_Gallery')) {
            return PEAR::raiseError('Galleries must be Ansel_Gallery objects or extend that class.');
        }
        return $this->_datatree->add($gallery);
    }

    /**
     * Return a list of users who have galleries with the given permissions
     * for the current user.
     *
     * @param integer $perm       The level of permissions required.
     * @param string $parent      The parent gallery to start looking in.
     * @param boolean $allLevels  Return all levels, or just the direct

     *                            children of $parent? Defaults to all levels.
     * @param integer $from       The gallery to start listing at.
     * @param integer $count      The number of galleries to return.
     *
     * @return array  List of users.
     */
    function listOwners($perm = PERMS_SHOW, $parent = DATATREE_ROOT, $allLevels = true, $from = 0, $count = 0)
    {
        $all_owners = $this->_datatree->getAttributeValues('owner', null, $parent, $allLevels);
        $owners = array();
        $item = 0;
        foreach ($all_owners as $owner) {
            if ($count) {
                $item++;
                if ($item <= $from) {
                    continue;
                }
                if ($item - $from > $count) {
                    continue;
                }
            }

            if ($this->countShares(Auth::getAuth(), $perm, $owner, $parent, $allLevels)) {
                $owners[] = $owner;
            }
        }
        return $owners;

    }

    /**
     * Count the number of users who have galleries with the given permissions
     * for the current user.
     *
     * @param integer $perm       The level of permissions required.
     * @param string $parent      The parent gallery to start looking in
     * @param boolean $allLevels  Return all levels, or just the direct
     *                            children of $parent? Defaults to all levels.
     *
     * @return integer  Number of users.
     */
    function countOwners($perm = PERMS_SHOW, $parent = DATATREE_ROOT, $allLevels = true)
    {
        return count($this->listOwners($perm, $parent, $allLevels));
    }

    /**
     * Return a list of categories containing galleries with the given
     * permissions for the current user.
     *
     * @param integer $perm   The level of permissions required.
     * @param integer $from   The gallery to start listing at.
     * @param integer $count  The number of galleries to return.
     *
     * @return array  List of categories.
     */
    function listCategories($perm = PERMS_SHOW, $from = 0, $count = 0)
    {
        $all_categories = Prefs_CategoryManager::get();
        $all_categories[] = '';

        $categories = array();
        $item = 0;
        foreach ($all_categories as $category) {
            if ($count) {
                $item++;
                if ($item <= $from) {
                    continue;
                }
                if ($item - $from > $count) {
                    continue;
                }
            }

            if ($this->countShares(Auth::getAuth(), $perm, array('category' => $category))) {
                $categories[] = $category;
            }
        }

        return $categories;
    }

    /**
     * Count the number of categories containing galleries with the given
     * permissions for the current user.
     *
     * @param integer $perm  The level of permissions required.
     *
     * @return integer  Number of categories.
     */
    function countCategories($perm = PERMS_SHOW)
    {
        return count($this->listCategories());
    }

    /**
     * Returns an Ansel_Gallery object corresponding to the given
     * share name, with the details retrieved appropriately.
     *
     * @param string $name  The name of the gallery to retrieve.
     *
     * @return Ansel_Gallery  The requested gallery.
     */
    function &getShare($name)
    {
        if (isset($this->_cache[$name])) {
            return $this->_cache[$name];
        }

        $this->_cache[$name] = &$this->_datatree->getObject($name, $this->_shareObject);
        if (!is_a($this->_cache[$name], 'PEAR_Error')) {
            $this->_cache[$name]->setShareOb($this);
            $this->_shareMap[$this->_cache[$name]->getId()] = $name;
        }

        return $this->_cache[$name];
    }

    /**
     * Returns an Ansel_Gallery object corresponding to the given
     * unique ID, with the details retrieved appropriately.
     *
     * @param string $cid  The id of the gallery to retrieve.
     *
     * @return Ansel_Gallery The requested gallery.
     */
    function &getShareById($cid)
    {
        if (isset($this->_shareMap[$cid])) {
            $share = &$this->_cache[$this->_shareMap[$cid]];
        } else {
            $share = $this->_datatree->getObjectById($cid, $this->_shareObject);
            if (!is_a($share, 'PEAR_Error')) {
                $share->setShareOb($this);
                $name = $share->getName();
                $this->_cache[$name] = &$share;
                $this->_shareMap[$cid] = $name;
            }
        }

        return $share;
    }

    /**
     * Returns an array of Ansel_Gallery objects corresponding to the
     * given set of unique IDs, with the details retrieved
     * appropriately.
     *
     * @param array $cids  The array of ids to retrieve.
     *
     * @return array  The requested shares.
     */
    function &getShares($cids)
    {
        $all_shares = array();
        $missing_ids = array();
        foreach ($cids as $cid) {
            if (isset($this->_shareMap[$cid])) {
                $all_shares[$this->_shareMap[$cid]] = &$this->_cache[$this->_shareMap[$cid]];
            } else {
                $missing_ids[] = $cid;
            }
        }

        if (count($missing_ids)) {
            $shares = $this->_datatree->getObjects($missing_ids, $this->_shareObject);
            if (is_a($shares, 'PEAR_Error')) {
                return $shares;
            }

            $keys = array_keys($shares);
            foreach ($keys as $key) {
                if (is_a($shares[$key], 'PEAR_Error')) {
                    return $shares[$key];
                }

                $shares[$key]->setShareOb($this);
                $all_shares[$key] = &$shares[$key];
                $this->_cache[$key] = &$shares[$key];
                $this->_shareMap[$shares[$key]->getId()] = $key;
            }
        }

        return $all_shares;
    }

    /**
     * Removes a share from the shares system permanently.
     *
     * @param Ansel_Gallery $share  The share to remove.
     *
     * @return boolean|PEAR_Error  PEAR_Error on failure.
     */
    function removeShare($share)
    {
        if (!is_a($share, 'Ansel_Gallery')) {
            return PEAR::raiseError('Invalid data.');
        }

        $result = Horde::callHook('_horde_hook_share_remove', array($share),
                                  'horde', false);
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        /* Remove share from the caches. */
        $id = $share->getId();
        unset($this->_shareMap[$id]);
        unset($this->_cache[$share->getName()]);

        /* Reset caches that depend on unknown criteria. */
        $this->_listCache = array();
        $this->_counts = array();

        return $this->_datatree->remove($share);
    }

    /**
     * Checks to see if a share has any child shares.
     *
     * @param Ansel_Gallery $share  The share to check for children.
     *
     * @return boolean  True if the specified share has child shares.
     */
    function hasChildren($share)
    {
        if (!is_a($share, 'Ansel_Gallery')) {
            return PEAR::raiseError('Invalid data');
        }

        $id = $share->getId();
        if (!isset($this->_hasChildren[$id])) {
            $this->_hasChildren[$id] = (bool)$this->_datatree->getNumberOfChildren($share);
        }

        return $this->_hasChildren[$id];
    }

    /**
     * Returns a share's direct parent object.
     *
     * @param string $share  Get the parent of this share.
     *
     * @return Ansel_Gallery The parent share, if it exists.
     */
    function &getParent($child)
    {
        $id = $child->getId();

        if (isset($this->_shareParents[$id])) {
            $parent = $this->_shareParents[$id];
        } else {
            $parent = $this->_datatree->getParent($child);
            if (is_a($parent, 'PEAR_Error')) {
                return $parent;
            }

            if (!$parent || ($parent == DATATREE_ROOT)) {
                return PEAR::raiseError('Parent does not exist.');
            }

            $this->_shareParents[$id] = $parent;
        }

        return $this->getShareById($parent);
    }

    /**
     * Utility function to be used with uasort() for sorting arrays of
     * Ansel_Share objects.
     * Example:<code>
     * uasort($list, array('Ansel_Share', '_sortShares'));
     * </code>
     *
     * @access private
     */
    function _sortShares($a, $b)
    {
        $aParts = explode(':', $a->getName());
        $bParts = explode(':', $b->getName());

        $min = min(count($aParts), count($bParts));
        $idA = '';
        $idB = '';
        for ($i = 0; $i < $min; $i++) {
            if ($idA) {
                $idA .= ':';
                $idB .= ':';
            }
            $idA .= $aParts[$i];
            $idB .= $bParts[$i];

            if ($idA != $idB) {
                $curA = isset($this->_sortList[$idA]) ? $this->_sortList[$idA]->get('name') : '';
                $curB = isset($this->_sortList[$idB]) ? $this->_sortList[$idB]->get('name') : '';
                return strnatcasecmp($curA, $curB);
            }
        }

        return count($aParts) > count($bParts);
    }

    /**
     * Checks if a share exists in the system.
     *
     * @param string $share  The share to check.
     *
     * @return boolean  True if the share exists, false otherwise.
     */
    function exists($share)
    {
        if (isset($this->_cache[$share])) {
            return true;
        }

        return $this->_datatree->exists($share);
    }

    /**
     * Returns the count of all shares that $userid has access to.
     *
     * @param string  $userid      The userid of the user to check access for.
     * @param integer $perm        The level of permissions required.
     * @param mixed   $attributes  Restrict the shares counted to those
     *                             matching $attributes. An array of
     *                             attribute/values pairs or a share owner
     *                             username.
     * @param string  $parent      The parent share to start searching at.
     * @param boolean $allLevels   Return all levels, or just the direct
     *                             children of $parent? Defaults to all levels.
     *
     * @return integer  Number of shares the user has access to.
     */
    function countShares($userid, $perm = PERMS_SHOW, $attributes = null,
                         $parent = DATATREE_ROOT, $allLevels = true)
    {
        $key = serialize(array($userid, $perm, $attributes, $parent, $allLevels));
        if (!isset($this->_counts[$key])) {
            $criteria = $this->_getShareCriteria($userid, $perm, $attributes);
            $this->_counts[$key] = $this->_datatree->countByAttributes($criteria, $parent, $allLevels, 'id');
        }

        return $this->_counts[$key];
    }

    /**
     * Returns an array of all shares that $userid has access to.
     *
     * @param string  $userid       The userid of the user to check access for.
     * @param integer $perm         The level of permissions required.
     * @param mixed   $attributes   Restrict the shares counted to those
     *                              matching $attributes. An array of
     *                              attribute/values pairs or a share owner
     *                              username.
     * @param string  $parent       The parent share to start searching at.
     * @param boolean $allLevels    Return all levels, or just the direct
     *                              children of $parent? Defaults to all
     *                              levels.
     * @param integer $from         The share to start listing at.
     * @param integer $count        The number of shares to return.
     * @param string  $sortby_name  Attribute name to use for sorting.
     * @param string  $sortby_key   Attribute key to use for sorting.
     * @param integer $direction    Sort direction:
     *                                0 - ascending
     *                                1 - descending
     *
     * @return array  The shares the user has access to.
     */
    function &listShares($userid, $perm = PERMS_SHOW, $attributes = null,
                         $parent = DATATREE_ROOT, $allLevels = true, $from = 0,
                         $count = 0, $sortby_name = null, $sortby_key = null,
                         $direction = 0)
    {
        $key = serialize(array($userid, $perm, $attributes, $parent,
                               $allLevels, $from, $count, $sortby_name,
                               $sortby_key, $direction));
        if (empty($this->_listCache[$key])) {
            $criteria = $this->_getShareCriteria($userid, $perm, $attributes);
            $sharelist = $this->_datatree->getByAttributes(
                $criteria, $parent, $allLevels, 'id', $from, $count,
                $sortby_name, $sortby_key, $direction);
            if (is_a($sharelist, 'PEAR_Error')) {
                return $sharelist;
            }
            $this->_listCache[$key] = array_keys($sharelist);
        }

        if (!count($this->_listCache[$key])) {
            return $this->_listCache[$key];
        }

        /* Make sure getShares() didn't return an error. */
        $shares = &$this->getShares($this->_listCache[$key]);
        if (is_a($shares, 'PEAR_Error')) {
            return $shares;
        }

        $this->_sortList = $shares;
        uasort($shares, array($this, '_sortShares'));
        $this->_sortList = null;

        $result = Horde::callHook('_horde_hook_share_list', array($userid,
                                  $perm, $attributes, $shares),
                                  'horde', false);
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        return $shares;
    }

    /**
     * TODO
     *
     * @see Perms::getPermissions
     *
     * @param TODO
     * @param TODO
     *
     * @return TODO
     */
    function getPermissions($share, $user = null)
    {
        if (!is_a($share, 'Ansel_Gallery')) {
            $share = &$this->getShare($share);
        }

        $perm = $share->getPermission();
        return $GLOBALS['perms']->getPermissions($perm, $user);
    }

    /**
     * Returns an array of criteria for querying shares.
     * @access private
     *
     * @param string  $userid      The userid of the user to check access for.
     * @param integer $perm        The level of permissions required.
     * @param mixed   $attributes  Restrict the shares returned to those who
     *                             have these attribute values.
     *
     * @return array  The criteria tree for fetching this user's shares.
     */
    function _getShareCriteria($userid, $perm = PERMS_SHOW, $attributes = null)
    {
        if (!empty($userid)) {
            $criteria = array(
                'OR' => array(
                    // (owner == $userid)
                    array(
                        'AND' => array(
                            array('field' => 'name', 'op' => '=', 'test' => 'owner'),
                            array('field' => 'value', 'op' => '=', 'test' => $userid))),

                    // (name == perm_users and key == $userid and val & $perm)
                    array(
                        'AND' => array(
                            array('field' => 'name', 'op' => '=', 'test' => 'perm_users'),
                            array('field' => 'key', 'op' => '=', 'test' => $userid),
                            array('field' => 'value', 'op' => '&', 'test' => $perm))),

                    // (name == perm_creator and val & $perm)
                    array(
                        'AND' => array(
                            array('field' => 'name', 'op' => '=', 'test' => 'perm_creator'),
                            array('field' => 'value', 'op' => '&', 'test' => $perm))),

                    // (name == perm_default and val & $perm)
                    array(
                        'AND' => array(
                            array('field' => 'name', 'op' => '=', 'test' => 'perm_default'),
                            array('field' => 'value', 'op' => '&', 'test' => $perm)))));

            // If the user has any group memberships, check for those also.
            require_once 'Horde/Group.php';
            $group = &Group::singleton();
            $groups = $group->getGroupMemberships($userid, true);
            if (is_array($groups) && count($groups)) {
                // (name == perm_groups and key in ($groups) and val & $perm)
                $criteria['OR'][] = array(
                    'AND' => array(
                        array('field' => 'name', 'op' => '=', 'test' => 'perm_groups'),
                        array('field' => 'key', 'op' => 'IN', 'test' => array_keys($groups)),
                        array('field' => 'value', 'op' => '&', 'test' => $perm)));
            }
        } else {
            $criteria = array(
                'AND' => array(
                     array('field' => 'name', 'op' => '=', 'test' => 'perm_guest'),
                     array('field' => 'value', 'op' => '&', 'test' => $perm)));
        }

        if (is_array($attributes)) {
            // Build attribute/key filter.
            foreach ($attributes as $key => $value) {
                $criteria = array(
                    'AND' => array(
                        $criteria,
                        array(
                            'JOIN' => array(
                                'AND' => array(
                                    array('field' => 'name', 'op' => '=', 'test' => $key),
                                    array('field' => 'value', 'op' => '=', 'test' => $value))))));
            }
        } elseif (!empty($attributes)) {
            // Restrict to shares owned by the user specified in the
            // $attributes string.
            $criteria = array(
                'AND' => array(
                    $criteria,
                    array(
                        'JOIN' => array(
                            array('field' => 'name', 'op' => '=', 'test' => 'owner'),
                            array('field' => 'value', 'op' => '=', 'test' => $attributes)))));
        }

        return $criteria;
    }

}

/**
 * Extension of the DataTreeObject class for storing Galleries.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @since   Horde 3.0
 * @package Ansel
 */
class Ansel_Gallery extends DataTreeObject {

    /**
     * The Ansel_Share object which this share came from - needed for
     * updating data in the backend to make changes stick, etc.
     *
     * @var Ansel_Share
     */
    var $_shareOb;

    /**
     * The Ansel_Gallery constructor. Calls the parent constructor
     * and sets the default image dimensions.
     *
     * @param string $name  The name of the gallery
     */
    function Ansel_Gallery($name)
    {
        parent::DataTreeObject($name);
        if (is_null($this->data)) {
            $this->data = array();
        }

        /* Initialize gallery info variables. */
        $this->data['default-type'] = 'auto';
        $this->data['date-created'] = time();
        $this->data['last-modified'] = time();
        $this->data['category'] = '';
    }

    /**
     * Associates a Share object with this share.
     *
     * @param Ansel_Share $shareOb  The Share object.
     */
    function setShareOb(&$shareOb)
    {
        $this->_shareOb = &$shareOb;
    }

    /**
     * Returns this share's parent object.
     *
     * @return Ansel_Gallery  The parent gallery, if it exists.
     */
    function &getParent()
    {
        $parent = $this->_shareOb->getParent($this);
        return $parent;
    }

    /**
     * Checks to see if a user has a given permission.
     *
     * @param string $userid       The userid of the user.
     * @param integer $permission  A PERMS_* constant to test for.
     * @param string $creator      The creator of the event.
     *
     * @return boolean  Whether or not $userid has $permission.
     */
    function hasPermission($userid, $permission, $creator = null)
    {
        if ($userid == $this->get('owner')) {
            return true;
        }

        return $GLOBALS['perms']->hasPermission($this->getPermission(), $userid, $permission, $creator);
    }

    /**
     * TODO
     *
     * @param TODO
     * @param boolean $update  TODO
     *
     * @return TODO
     */
    function setPermission(&$perm, $update = true)
    {
        $this->data['perm'] = $perm->getData();
        if ($update) {
            return $this->save();
        }
        return true;
    }

    /**
     * TODO
     *
     * @return TODO
     */
    function getPermission()
    {
        $perm = new DataTreeObject_Permission($this->getName());
        $perm->data = isset($this->data['perm']) ? $this->data['perm'] : array();

        return $perm;
    }

    /**
     * Forces all children of this share to inherit the permissions set on this
     * share.
     *
     * @return TODO
     */
    function inheritPermissions()
    {
        $c_list = $this->datatree->get(DATATREE_FORMAT_FLAT, $this->getName(), true);
        if (is_a($c_list, 'PEAR_Error') || !$c_list) {
            // If we got back an error or an empty array, just return it.
            return $c_list;
        }
        unset($c_list[$this->getName()]);

        $children = &$this->_shareOb->getShares(array_keys($c_list));
        if (is_a($children, 'PEAR_Error')) {
            return $children;
        }

        $perm = $this->getPermission();
        foreach ($children as $child) {
            $child->setPermission($perm);
        }

        return true;
    }

    /**
     * Saves any changes to this object to the backend permanently. New objects
     * are added instead.
     *
     * @return boolean | PEAR_Error  PEAR_Error on failure.
     */
    function save()
    {
        $result = Horde::callHook('_horde_hook_share_modify', array($this),
                                  'horde', false);
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        return parent::save();
    }

    /**
     * Add an image to this gallery.
     *
     * @param array $image_data  The image to add. Required keys include
     *                           'image_caption', and 'data'. Optional keys include
     *                           'image_filename' and 'image_type'
     *
     * @return  The datatree name for the new image.
     */
    function addImage($image_data)
    {
        global $conf;

        if (!isset($image_data['image_filename'])) {
            $image_data['image_filename'] = 'Untitled';
        }
        $image_data['gallery_id'] = $this->getId();
        $image_data['image_sort'] = $this->count();

        $image = new Ansel_Image($image_data);
        $result = $image->save();
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        if (!empty($image_data['data'])) {
            $result = $image->writeData();
            if (is_a($result, 'PEAR_Error')) {
                return $result;
            }
        }

        require_once ANSEL_BASE . '/lib/Exif.php';
        $exif_fields = Ansel_ImageData::getExifData($image);

        /* Auto-rotate the image if there is appropriate Exif
         * information based on which to do so. */
        if (isset($exif_fields['orientation']) && $exif_fields['orientation'] != 1) {
            switch ($exif_fields['orientation']) {
            case 2:
                $image->mirror();
                break;

            case 3:
                $image->rotate(180);
                break;

            case 4:
                $image->mirror();
                $image->rotate(180);
                break;

            case 5:
                $image->flip();
                $image->rotate(90);
                break;

            case 6:
                $image->rotate(90);
                break;

            case 7:
                $image->mirror();
                $image->rotate(90);
                break;

            case 8:
                $image->rotate(270);
                break;
            }

            $exif_fields['orientation'] = 1;
            $image->updateData($image->raw());
        }

        /* Save attributes. */
        $insert = $GLOBALS['ansel_db']->prepare('INSERT INTO ansel_image_attributes (image_id, attr_name, attr_value) VALUES (?, ?, ?)');
        if (is_a($insert, 'PEAR_Error')) {
            return $insert;
        }
        foreach ($exif_fields as $name => $value) {
            $result = $insert->execute(array($image->id, $name, $value));
            if (is_a($result, 'PEAR_Error')) {
                return $result;
            }
        }

        /* Change the default image if it's set to automatic. */
        if ($this->data['default-type'] == 'auto') {
            $this->data['default'] = $image->id;
        }

        /* Update the modified flag and save gallery changes */
        $this->set('last-modified', time(), true);

        /* Return the ID of the new image. */
        return $image->id;
    }

    /**
     * Move images from this gallery to a new gallery.
     *
     * @param array $images An array of image ids.
     * @param Ansel_Gallery $gallery The gallery to move the images to.
     *
     * @return integer | PEAR_Error The number of images copied, or an error message.
     */
    function moveImagesTo($images, $gallery)
    {
        if (!$gallery->hasPermission(Auth::getAuth(), PERMS_EDIT)) {
            return PEAR::raiseError(sprintf(_("You are not allowed to move images to %s."), $newGallery->get('name')));
        } elseif (!$this->hasPermission(Auth::getAuth(), PERMS_DELETE)) {
            return PEAR::raiseError(sprintf(_("You are not allowed to remove images from %s."), $gallery->get('name')));
        }

        // Sanitize image ids, and see if we're removing our default
        // image.
        $ids = array();
        foreach ($images as $imageId) {
            $ids[] = (int)$imageId;

            if ($imageId == $this->data['default']) {
                $this->set('default', null, true);
            }
        }

        return $GLOBALS['ansel_db']->exec('UPDATE ansel_images SET gallery_id = ' . (int)$gallery->getId() . ' WHERE image_id IN (' . implode(',', $ids) . ')');
    }

    /**
     * Set the order of an image in this gallery.
     *
     * @param integer $imageId The image to sort.
     * @param integer $pos     The sort position of the image.
     */
    function setImageOrder($imageId, $pos)
    {
        return $GLOBALS['ansel_db']->exec('UPDATE ansel_images SET image_sort = ' . (int)$pos . ' WHERE image_id = ' . (int)$imageId . ' AND gallery_id = ' . (int)$this->getId());
    }

    /**
     * Remove the given image from this gallery.
     *
     * @param mixed   $image   Image to delete. Can be an Ansel_Image
     *                         or an image ID.
     *
     * @return boolean  True on success, false on failure.
     */
    function removeImage($image)
    {
        /* Make sure $image is an Ansel_Image; if not, try loading
         * it. */
        if (!is_a($image, 'Ansel_Image')) {
            $img = $this->getImage($image);
            if (is_a($img, 'PEAR_Error')) {
                return $img;
            }
            $image = $img;
        } else {
            /* Make sure the image is in this gallery. */
            if (is_a('PEAR_Error', $this->getImage($image->id))) {
                return false;
            }
        }

        /* Change gallery info. */
        if ($this->data['default'] == $image->id) {
            $this->data['default'] = null;
            $this->data['default-type'] = 'auto';
        }

        /* Update the modified flag and save gallery changes */
        $this->set('last-modified', time(), true);

        /* Delete cached files from VFS. */
        $image->deleteCache();

        /* Delete original image from VFS. */
        $GLOBALS['ansel_vfs']->deleteFile($image->vfspath('full'), $image->getVFSName('full'));

        /* Delete from SQL. */
        $GLOBALS['ansel_db']->exec('DELETE FROM ansel_images WHERE image_id = ' . (int)$image->id);

        return true;
    }

    /**
     * Set one of the attributes of the object.
     *
     * @param string  $attribute  The attribute to set.
     * @param mixed   $value      The value for $attribute.
     * @param boolean $update     Determines whether the backend should
     *                            be updated or not
     */
    function set($attribute, $value, $update = false)
    {
        parent::set($attribute, $value);
        if ($update) {
            $this->_shareOb->_datatree->updateData($this);
        }
    }

    /**
     * Returns the parents of this gallery in flat array.
     */
    function getParents()
    {
        $parents = array();
        $gallery = $this;
        while (is_a($gallery = $gallery->getParent(), 'Ansel_Gallery')) {
            $parents[] = $gallery->getId();
        }
        return array_reverse($parents);
    }

    /**
     * Returns the owner of this gallery.
     *
     * @return Identity object for the owner of this gallery.
     */
    function getOwner()
    {
        require_once 'Horde/Identity.php';
        $identity = &Identity::singleton('none', $this->get('owner'));
        return $identity;
    }

    /**
     * Lists a slice of the image ids in this gallery.
     *
     * @param integer $from  The image to start listing.
     * @param integer $count The numer of images to list.
     */
    function listImages($from = 0, $count = 0)
    {
        $GLOBALS['ansel_db']->setLimit($count, $from);
        return $GLOBALS['ansel_db']->queryCol('SELECT image_id FROM ansel_images WHERE gallery_id = ' . (int)$this->getId() . ' ORDER BY image_sort');
    }

    /**
     * Gets a slice of the images in this gallery.
     *
     * @param integer $from  The image to start fetching.
     * @param integer $count The numer of images to return.
     */
    function getImages($from = 0, $count = 0)
    {
        $GLOBALS['ansel_db']->setLimit($count, $from);
        $images = $GLOBALS['ansel_db']->queryAll('SELECT image_id, gallery_id, image_filename, image_type, image_caption, image_uploaded, image_sort FROM ansel_images WHERE gallery_id = ' . (int)$this->getId() . ' ORDER BY image_sort', null, MDB2_FETCHMODE_ASSOC);
        if (is_null($images)) {
            return PEAR::raiseError('not found');
        } elseif (is_a($images, 'PEAR_Error')) {
            return $images;
        }

        $objects = array();
        foreach ($images as $image) {
            $objects[] = new Ansel_Image($image);
        }
        return $objects;
    }

    /**
     * Returns the image in this gallery corresponding to the given id.
     *
     * @param integer $id  The ID of the image to retrieve.
     *
     * @return Ansel_Image  The image object corresponding to the given name.
     */
    function getImage($id)
    {
        $image = $GLOBALS['ansel_db']->queryRow('SELECT image_id, gallery_id, image_filename, image_type, image_caption, image_uploaded, image_sort FROM ansel_images WHERE image_id = ' . (int)$id . ' AND gallery_id = ' . (int)$this->getId(), null, MDB2_FETCHMODE_ASSOC);
        if (is_null($image)) {
            return PEAR::raiseError('not found');
        } elseif (is_a($image, 'PEAR_Error')) {
            return $image;
        } else {
            return new Ansel_Image($image);
        }
    }

    /**
     * Returns the number of images in this gallery and, optionally, all
     * sub-galleries.
     *
     * @param boolean $subgalleries  Determines whether subgalleries should
     *                               be counted or not.
     *
     * @return integer number of images in this gallery
     */
    function count($subgalleries = false)
    {
        if ($subgalleries) {
            $count = $this->count();
            $galleries = $this->_shareOb->listShares(Auth::getAuth(), PERMS_SHOW, '', $this->getName());
            foreach ($galleries as $galleryId => $gallery) {
                $count += $gallery->count();
            }
            return $count;
        }

        return $GLOBALS['ansel_db']->queryOne('SELECT COUNT(*) FROM ansel_images WHERE gallery_id = '. (int)$this->getId());
    }

    /**
     * Returns the default image for this gallery.
     */
    function getDefaultImage()
    {
        // BC check for when we initialized the default image to -1.
        if (!empty($this->data['default']) && $this->data['default'] > 0) {
            return $this->data['default'];
        }

        if ($this->count()) {
            $keys = $this->listImages();
            $this->set('default', $keys[count($keys) - 1]);
            $this->set('default-type', 'auto', true);
            return $this->data['default'];
        }

        $galleries = $this->_shareOb->listShares(Auth::getAuth(), PERMS_SHOW, '', $this->getName());
        if ($galleries) {
            foreach ($galleries as $galleryId => $gallery) {
                if ($default_img = $gallery->getDefaultImage()) {
                    return $default_img;
                }
            }
        }
    }

    /**
     * Maps this object's attributes from the data array into a format that we
     * can store in the attributes storage backend.
     *
     * @access protected
     *
     * @return array  The attributes array.
     */
    function _toAttributes()
    {
        // Default to no attributes.
        $attributes = array();

        foreach ($this->data as $key => $value) {
            if ($key == 'perm') {
                foreach ($value as $type => $perms) {
                    if (is_array($perms)) {
                        foreach ($perms as $member => $perm) {
                            $attributes[] = array('name' => 'perm_' . $type,
                                                  'key' => $member,
                                                  'value' => $perm);
                        }
                    } else {
                        $attributes[] = array('name' => 'perm_' . $type,
                                              'key' => '',
                                              'value' => $perms);
                    }
                }
            } else {
                $attributes[] = array('name' => $key,
                                      'key' => '',
                                      'value' => $value);
            }
        }

        return $attributes;
    }

    /**
     * Takes in a list of attributes from the backend and maps it to our
     * internal data array.
     *
     * @access protected
     *
     * @param array $attributes   The list of attributes from the backend
     *                            (attribute name, key, and value).
     */
    function _fromAttributes($attributes)
    {
        // Initialize data array.
        $this->data['perm'] = array();

        foreach ($attributes as $attr) {
            if (substr($attr['name'], 0, 4) == 'perm') {
                if (!empty($attr['key'])) {
                    $this->data['perm'][substr($attr['name'], 5)][$attr['key']] = $attr['value'];
                } else {
                    $this->data['perm'][substr($attr['name'], 5)] = $attr['value'];
                }
            } else {
                $this->data[$attr['name']] = $attr['value'];
            }
        }
    }

}

/**
 * @package Ansel
 */
class Ansel_Image {

    var $id = null;
    var $filename = 'Untitled';
    var $caption = '';
    var $type = 'image/jpeg';
    var $uploaded;
    var $sort;
    var $gallery;

    var $_image;

    var $_dirty;
    var $_loaded = array();
    var $_data = array();

    function Ansel_Image($image = array())
    {
        if ($image) {
            $this->filename = $image['image_filename'];
            $this->caption = $image['image_caption'];
            $this->sort = $image['image_sort'];
            $this->gallery = $image['gallery_id'];

            if (!empty($image['image_uploaded'])) {
                $this->uploaded = $image['image_uploaded'];
            }

            if (!empty($image['image_id'])) {
                $this->id = $image['image_id'];
            }

            if (!empty($image['image_type'])) {
                $this->type = $image['image_type'];
            }

            if (!empty($image['data'])) {
                $this->_data['full'] = $image['data'];
            }
        }

        $this->_image = Ansel::getImageObject();
        $this->_image->reset();
    }

    function meta($key)
    {
    }

    /**
     * Return the vfs path for this image.
     *
     * @return string  The vfs path for this image.
     */
    function vfspath($view = 'full')
    {
        return '.horde/ansel/' . substr($this->id, -2) . '/' . $view;
    }

    /**
     * Returns the file name of this image as used in the VFS backend.
     *
     * @return string  This image's VFS file name.
     */
    function getVFSName($view)
    {
        $vfsname = $this->id;

        require_once 'Horde/MIME/Magic.php';
        if ($view == 'full' && $this->type) {
            $type = $this->type;
        } else {
            $type = 'image/' . $GLOBALS['conf']['image']['type'];
        }
        if ($ext = MIME_Magic::MIMEToExt($type)) {
            $vfsname .= '.' . $ext;
        }

        return $vfsname;
    }

    /**
     * Loads the given view into memory.
     *
     * @param string $view  Which view to load.
     */
    function load($view = 'full')
    {
        /* If we've already loaded the data, just return now. */
        if (!empty($this->_loaded[$view])) {
            return;
        }

        $result = $this->createView($view);
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        /* If createView() had to resize the full image, we've already
         * loaded the data, so return now. */
        if (!empty($this->_loaded[$view])) {
            return;
        }

        /* We've definitely successfully loaded the image now. */
        $this->_loaded[$view] = true;

        /* Get the VFS info. */
        $vfspath = $this->vfspath($view);
        if (is_a($vfspath, 'PEAR_Error')) {
            return $vfspath;
        }

        /* Read in the requested view. */
        $data = $GLOBALS['ansel_vfs']->read($vfspath, $this->getVFSName($view));
        if (is_a($data, 'PEAR_Error')) {
            return $data;
        }

        $this->_data[$view] = $data;
        $this->_image->loadString($vfspath . '/' . $this->id, $data);
    }

    /**
     * Creates and caches the given view.
     *
     * @param string view  Which view to create.
     */
    function createView($view)
    {
        /* Get the VFS info. */
        $vfspath = $this->vfspath($view);
        if (is_a($vfspath, 'PEAR_Error')) {
            return $vfspath;
        }

        if ($GLOBALS['ansel_vfs']->exists($vfspath, $this->getVFSName($view))) {
            return;
        }

        $data = $GLOBALS['ansel_vfs']->read($this->vfspath('full'), $this->getVFSName('full'));
        if (is_a($data, 'PEAR_Error')) {
            return $data;
        }
        if (is_a($result = $this->_image->loadString($this->vfspath('full') . '/' . $this->id, $data), 'PEAR_Error')) {
            return $result;
        }

        switch ($view) {
        case 'screen':
            /* Create and cache the screen size image. */
            $dimensions = $this->_image->getDimensions();
            $this->_image->resize(min($GLOBALS['conf']['screen']['width'], $dimensions['width']),
                                  min($GLOBALS['conf']['screen']['height'], $dimensions['height']),
                                  true);

            /* Store the data, since we have it loaded now. */
            $this->_data[$view] = $this->_image->raw();
            $this->_image->loadString($vfspath . '/' . $this->id, $this->_data[$view]);
            $this->_loaded[$view] = true;

            /* Write the resized image to the backend. */
            $GLOBALS['ansel_vfs']->writeData($vfspath, $this->getVFSName($view), $this->_data[$view], true);
            break;

        case 'thumb':
            /* Create and cache the thumbnail. */
            $dimensions = $this->_image->getDimensions();
            $this->_image->resize(min($GLOBALS['conf']['thumbnail']['width'], $dimensions['width']),
                                  min($GLOBALS['conf']['thumbnail']['height'], $dimensions['height']),
                                  true);

            /* Store the data, since we have it loaded now. */
            $this->_data[$view] = $this->_image->raw();
            $this->_image->loadString($vfspath . '/' . $this->id, $this->_data[$view]);
            $this->_loaded[$view] = true;

            /* Write the resized image to the backend. */
            $GLOBALS['ansel_vfs']->writeData($vfspath, $this->getVFSName($view), $this->_data[$view], true);
            break;

        case 'mini':
            /* Create and cache the mini. */
            $dimensions = $this->_image->getDimensions();
            $this->_image->resize(min(50, $dimensions['width']), min(50, $dimensions['height']), true);

            /* Store the data, since we have it loaded now. */
            $this->_data[$view] = $this->_image->raw();
            $this->_image->loadString($vfspath . '/' . $this->id, $this->_data[$view]);
            $this->_loaded[$view] = true;

            /* Write the resized image to the backend. */
            $GLOBALS['ansel_vfs']->writeData($vfspath, $this->getVFSName($view), $this->_data[$view], true);
            break;
        }
    }

    /**
     * Writes the current data to vfs.  Any cached images are deleted
     * so they can be regenerated from the new data when necessary.
     */
    function writeData()
    {
        /* Delete cache and write data. */
        $this->deleteCache();
        $this->_dirty = false;

        return $GLOBALS['ansel_vfs']->writeData($this->vfspath('full'), $this->getVFSName('full'), $this->_data['full'], true);
    }

    /**
     * Save basic image details
     */
    function save()
    {
        if ($this->id) {
            // Update an existing image.
            $update = $GLOBALS['ansel_db']->prepare('UPDATE ansel_images SET image_filename = ?, image_type = ?, image_caption = ?, image_sort = ? WHERE image_id = ?');
            if (is_a($update, 'PEAR_Error')) {
                return $update;
            }
            return $update->execute(array($this->filename, $this->type, $this->caption, $this->sort, $this->id));
        }

        if (!$this->gallery || !strlen($this->filename) || !$this->type) {
            return PEAR::raiseError('Incomplete image');
        }

        // Saving a new image.
        $image_id = $GLOBALS['ansel_db']->nextId('ansel_images');
        if (is_a($image_id, 'PEAR_Error')) {
            return $image_id;
        }

        $insert = $GLOBALS['ansel_db']->prepare('INSERT INTO ansel_images (image_id, gallery_id, image_filename, image_type, image_caption, image_uploaded, image_sort) VALUES (?, ?, ?, ?, ?, ?, ?)');
        if (is_a($insert, 'PEAR_Error')) {
            return $insert;
        }

        $result = $insert->execute(array($image_id, $this->gallery, $this->filename, $this->type, $this->caption, time(), $this->sort));
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        $this->id = $image_id;
        return $this->id;
    }

    /**
     * Deletes the specified cache file.
     *
     * If none is specified, deletes all of the cache files.
     *
     * @param string $view  Which cache file to delete.
     */
    function deleteCache($view = 'all')
    {
        /* Delete cached screen image. */
        if ($view == 'all' || $view == 'screen') {
            $GLOBALS['ansel_vfs']->deleteFile($this->vfspath('screen'), $this->getVFSName('screen'));
        }

        /* Delete cached thumbnail. */
        if ($view == 'all' || $view == 'thumb') {
            $GLOBALS['ansel_vfs']->deleteFile($this->vfspath('thumb'), $this->getVFSName('thumb'));
        }

        /* Delete cached mini image. */
        if ($view == 'all' || $view == 'mini') {
            $GLOBALS['ansel_vfs']->deleteFile($this->vfspath('mini'), $this->getVFSName('mini'));
        }
    }

    /**
     * Returns the raw data for the given view.
     *
     * @param string $view  Which view to return.
     */
    function raw($view = 'full')
    {
        $this->load($view);
        if ($this->_dirty) {
            return $this->_image->raw();
        } else {
            return $this->_data[$view];
        }
    }

    /**
     * Change the image data. Deletes old cache and writes the new
     * data to the VFS.
     *
     * @param string $data  The new data for this image.
     */
    function updateData($data)
    {
        if (is_a($data, 'PEAR_Error')) {
            return $data;
        }

        /* Delete old cached data and write the full data. */
        $this->deleteCache();
        return $GLOBALS['ansel_vfs']->writeData($this->vfspath('full'), $this->getVFSName('full'), $data, true);
    }

    /**
     * Sends the correct HTTP headers to the browser to download this image.
     *
     * @param string $view  The view to download.
     */
    function downloadHeaders($view = 'full')
    {
        global $browser, $conf;

        $filename = $this->filename;
        if ($view != 'full') {
            require_once 'Horde/MIME/Magic.php';
            if ($ext = MIME_Magic::MIMEToExt('image/' . $conf['image']['type'])) {
                $filename .= '.' . $ext;
            }
        }

        $browser->downloadHeaders($filename);
    }

    /**
     * Display the requested view.
     *
     * @param string $view  Which view to display.
     */
    function display($view = 'full')
    {
        if ($view == 'full' && !$this->_dirty) {
            $data = $GLOBALS['ansel_vfs']->read($this->vfspath('full'), $this->getVFSName('full'));
            if (is_a($data, 'PEAR_Error')) {
                return $data;
            }
            echo $data;
            return;
        }
        if (is_a($result = $this->load($view), 'PEAR_Error')) {
            return $result;
        }
        return $this->_image->display();
    }

    /**
     * Wraps the given fiew into a file.
     *
     * @param string $view  Which view to wrap up.
     */
    function toFile($view = 'full')
    {
        if (is_a(($result = $this->load($view)), 'PEAR_Error')) {
            return $result;
        }
        return $this->_image->toFile($this->_dirty ? false : $this->_data[$view]);
    }

    /**
     * Returns the dimensions of the given view.
     *
     * @param string $view  The view (size) to check dimensions for.
     */
    function getDimensions($view = 'full')
    {
        if (is_a(($result = $this->load($view)), 'PEAR_Error')) {
            return $result;
        }
        return $this->_image->getDimensions();
    }

    /**
     * Rotates the image.
     *
     * @param string $view The view (size) to work with.
     * @param integer $angle  What angle to rotate the image by.
     */
    function rotate($view = 'full', $angle)
    {
        $this->load($view);
        $this->_dirty = true;
        return $this->_image->rotate($angle);
    }

    /**
     * Converts the image to grayscale.
     *
     * @param string $view The view (size) to work with.
     */
    function grayscale($view = 'full')
    {
        $this->load($view);
        $this->_dirty = true;
        return $this->_image->grayscale();
    }

    /**
     * Watermarks the image.
     *
     * @param string $view The view (size) to work with.
     * @param string $watermark  String to use as the watermark.
     */
    function watermark($view = 'full', $watermark = null, $halign = null, $valign = null, $font = null)
    {
        if (empty($watermark)) {
            require_once 'Horde/Identity.php';
            $identity = &Identity::singleton();
            $name = $identity->getValue('fullname');
            if (empty($name)) {
                $name = Auth::getAuth();
            }
            $watermark = sprintf(_("(c) %s %s"), date('Y'), $name);
        }

        $this->load($view);
        $this->_dirty = true;
        return $this->_image->watermark($watermark, $halign, $valign, $font);
    }

    /**
     * Flips the image.
     *
     * @param string $view The view (size) to work with.
     */
    function flip($view = 'full')
    {
        $this->load($view);
        $this->_dirty = true;
        return $this->_image->flip();
    }

    /**
     * Mirrors the image.
     *
     * @param string $view The view (size) to work with.
     */
    function mirror($view = 'full')
    {
        $this->load($view);
        $this->_dirty = true;
        return $this->_image->mirror();
    }

}
