<?php
/**
 * Library for extracting EXIF information from digital photos.
 *
 * $Horde: ansel/lib/Exif.php,v 1.36 2007/01/02 00:40:24 jan Exp $
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 *
 * @package Ansel
 */
class Ansel_ImageData {

    /**
     * Get the image attributes from the datatree backend.
     *
     * @param Ansel_Image $image  The image to retrieve attributes for.
     *                             attributes for.
     *
     * @return array  The formatted output of the EXIF data.
     */
    function getAttributes($image)
    {
        $attributes = $GLOBALS['ansel_db']->queryAll('SELECT attr_name, attr_value FROM ansel_image_attributes WHERE image_id = ' . (int)$image->id, null, MDB2_FETCHMODE_ASSOC, true);
        $fields = Ansel_ImageData::getFields();
        $output = array();

        foreach ($fields as $field => $data) {
            if (!isset($attributes[$field])) {
                continue;
            }

            $value = Ansel_ImageData::getHumanReadable($field, $attributes[$field]);
            $description = isset($data['description']) ? $data['description'] : $field;
            $output[] = '<td><strong>' . $description . '</strong></td><td>' . htmlspecialchars($value) . '</td>';
        }

        return $output;
    }

    /**
     * Get the EXIF data from an image, process it, and return it.
     *
     * @see getFields()
     *
     * @param Ansel_Image $image  The image to read exif data from.
     * @return array Array of EXIF attributes.
     */
    function getExifData($image)
    {
        // Unfortunately, the PHP function requires a file, not a
        // stream.
        $imageFile = $image->toFile();

        if (function_exists('exif_read_data')) {
            $exif = @exif_read_data($imageFile, 0, false);
        } else {
            $raw = read_exif_data_raw($imageFile);
            $exif = array();
            foreach ($raw as $key => $value) {
                if (($key == 'IFD0') || ($key == 'SubIFD')) {
                    foreach ($value as $subkey => $subvalue) {
                        $exif[$subkey] = $subvalue;
                    }
                } else {
                    $exif[$key] = $value;
                }
            }
        }

        // Make sure to clean up the temp file.
        @unlink($imageFile);

        // See if we got any attributes back.
        $results = array();
        if ($exif) {
            $fields = Ansel_ImageData::getFields();
            foreach ($fields as $field => $data) {
                $value = isset($exif[$field]) ? $exif[$field] : '';

                // Don't store empty fields.
                if ($value === '') {
                    continue;
                }

                // If the field is a date field, convert the value to a
                // timestamp.
                if ($data['type'] == 'date') {
                    @list($ymd, $hms) = explode(' ', $value, 2);
                    @list($year, $month, $day) = explode(':', $ymd, 3);
                    if (!empty($hms) && !empty($year) && !empty($month) && !empty($day)) {
                        $time = "$month/$day/$year $hms";
                        $value = strtotime($time);
                    }
                }

                $results[$field] = $value;
            }
        }

        return $results;
    }

    /**
     * Convert an exif field into human-readable form.
     *
     * @param string $field  The name of the field to translate.
     * @param string $data   The data value to translate.
     *
     * @return string  The converted data.
     */
    function getHumanReadable($field, $data)
    {
        switch ($field) {
        case 'ExposureMode':
            switch ($data) {
            case 0:
                return _("Easy shooting");
            case 1:
                return _("Program");
            case 2:
                return _("Tv-priority");
            case 3:
                return _("Av-priority");
            case 4:
                return _("Manual");
            case 5:
                return _("A-DEP");
            }
            break;

        case 'XResolution':
        case 'YResolution':
            if (strpos($data, '/') !== false) {
                list($n, $d) = explode('/', $data, 2);
                return sprintf(_("%d pixels"), $n);
            }
            break;

        case 'ExifImageWidth':
        case 'ExifImageLength':
            return sprintf(_("%d pixels"), $data);

        case 'Orientation':
            switch ($data) {
            case 1:
                return sprintf(_("Normal (O deg)"));
            case 2:
                return sprintf(_("Mirrored"));
            case 3:
                return sprintf(_("Upsidedown"));
            case 4:
                return sprintf(_("Upsidedown Mirrored"));
            case 5:
                return sprintf(_("90 deg CW Mirrored"));
            case 6:
                return sprintf(_("90 deg CCW"));
            case 7:
                return sprintf(_("90 deg CCW Mirrored"));
            case 8:
                return sprintf(_("90 deg CW"));
            }
            break;

        case 'ExposureTime':
            return sprintf(_("%d sec"), $data);

        case 'ShutterSpeedValue':
            if (strpos($data, '/') !== false) {
                list($n, $d) = explode('/', $data, 2);
                $data = $n / $d;
                $data = exp($data * log(2));
                if ($data > 1) {
                    $data = floor($data);
                }
                if ($data < 1) {
                    $data = round(1 / $data);
                } else {
                    $data = '1/' . $data;
                }
                return sprintf(_("%d sec"), $data);
            }
            break;

        case 'ApertureValue':
        case 'MaxApertureValue':
            if (strpos($data, '/') !== false) {
                list($n, $d) = explode('/', $data, 2);
                $data = $n / $d;
                $data = exp(($data * log(2)) / 2);

                // Precision is 1 digit.
                $data = round($data, 1);
                return 'f ' . $data;
            }
            break;

        case 'FocalLength':
            if (strpos($data, '/') !== false) {
                list($n, $d) = explode('/', $data, 2);
                return $n . ' mm';
            }
            break;

        case 'FNumber':
            if (strpos($data, '/') !== false) {
                list($n, $d) = explode('/', $data, 2);
                if ($d != 0) {
                    return 'f ' . round($n / $d, 1);
                }
            }
            break;

        case 'ExposureBiasValue':
            if (strpos($data, '/') !== false) {
                list($n, $d) = explode('/', $data, 2);
                if ($n == 0) {
                    return '0 EV';
                } else {
                    return $data . ' EV';
                }
            }
            break;

        case 'MeteringMode':
            switch ($data) {
            case 0: return _("Unknown");
            case 1: return _("Average");
            case 2: return _("Center Weighted Average");
            case 3: return _("Spot");
            case 4: return _("Multi-Spot");
            case 5: return _("Multi-Segment");
            case 6: return _("Partial");
            case 255: return _("Other");
            }
            break;

        case 'WhiteBalance':
            switch ($data) {
            case 0: return 'Auto';
            case 1: return _("Sunny");
            case 2: return _("Cloudy");
            case 3: return _("Tungsten");
            case 4: return _("Fluorescent");
            case 5: return _("Flash");
            case 6: return _("Custom");
            case 129: return _("Manual");
            }
            break;

        case 'FocalLength':
            return $data . ' mm';

        case 'FocalLengthIn35mmFilm':
            return $data . ' mm';

        case 'Flash':
            switch ($data) {
            case 0: return _("No Flash");
            case 1: return _("Flash");
            case 5: return _("Flash, strobe return light not detected");
            case 7: return _("Flash, strobe return light detected");
            case 9: return _("Compulsory Flash");
            case 13: return _("Compulsory Flash, Return light not detected");
            case 15: return _("Compulsory Flash, Return light detected");
            case 16: return _("No Flash");
            case 24: return _("No Flash");
            case 25: return _("Flash, Auto-Mode");
            case 29: return _("Flash, Auto-Mode, Return light not detected");
            case 31: return _("Flash, Auto-Mode, Return light detected");
            case 32: return _("No Flash");
            case 65: return _("Red Eye");
            case 69: return _("Red Eye, Return light not detected");
            case 71: return _("Red Eye, Return light detected");
            case 73: return _("Red Eye, Compulsory Flash");
            case 77: return _("Red Eye, Compulsory Flash, Return light not detected");
            case 79: return _("Red Eye, Compulsory Flash, Return light detected");
            case 89: return _("Red Eye, Auto-Mode");
            case 93: return _("Red Eye, Auto-Mode, Return light not detected");
            case 95: return _("Red Eye, Auto-Mode, Return light detected");
            }
            break;

        case 'FileSize':
            return sprintf(_("%d bytes"), $data);

        case 'SensingMethod':
            switch ($data) {
            case 1: return _("Not defined");
            case 2: return _("One Chip Color Area Sensor");
            case 3: return _("Two Chip Color Area Sensor");
            case 4: return _("Three Chip Color Area Sensor");
            case 5: return _("Color Sequential Area Sensor");
            case 7: return _("Trilinear Sensor");
            case 8: return _("Color Sequential Linear Sensor");
            }
            break;

        case 'ColorSpace':
            switch ($data) {
            case 1:
                return _("sRGB");
                break;
            default:
                return _("Uncalibrated");
            }

        case 'DateTime':
        case 'DateTimeOriginal':
        case 'DateTimeDigitized':
            return date('m/d/Y h:i:s O', $data);

        default:
            return $data;
        }
    }

    /**
     * Get the list of Exif fields we support, and their descriptions.
     *
     * @static
     *
     * @return array  Hash of fieldname => description.
     */
    function getFields()
    {
        return array('Make' => array('description' => _("Camera Make"), 'type' => 'text'),
                     'Model' => array('description' => _("Camera Model"), 'type' => 'text'),
                     'ImageType' => array('description' => _("Image Type"), 'type' => 'text'),
                     'ImageDescription' => array('description' => _("Image Description"), 'type' => 'text'),
                     'FileSize' => array('description' => _("File Size"), 'type' => 'number'),
                     'DateTime' => array('description' => _("Date Image Modified"), 'type' => 'date'),
                     'DateTimeOriginal' => array('description' => _("Date Image Taken"), 'type' => 'date'),
                     'DateTimeDigitized' => array('description' => _("Date Image Digitized"), 'type' => 'date'),
                     'ExifImageWidth' => array('description' => _("Width"), 'type' => 'number'),
                     'ExifImageLength' => array('description' => _("Height"), 'type' => 'number'),
                     'xResolution' => array('description' => _("X Resolution"), 'type' => 'number'),
                     'yResolution' => array('description' => _("Y Resolution"), 'type' => 'number'),
                     'ShutterSpeedValue' => array('description' => _("Shutter Speed"), 'type' => 'number'),
                     'ExposureTime' => array('description' => _("Exposure"), 'type' => 'number'),
                     'FocalLength' => array('description' => _("Focal Length"), 'type' => 'number'),
                     'FocalLengthIn35mmFilm' => array('description' => _("Focal Length (35mm equiv)"), 'type' => 'number'),
                     'ApertureValue' => array('description' => _("Aperture"), 'type' => 'number'),
                     'FNumber' => array('description' => _("F-Number"), 'type' => 'number'),
                     'ISOSpeedRatings' => array('description' => _("ISO Setting"), 'type' => 'number'),
                     'ExposureBiasValue' => array('description' => _("Exposure Bias"), 'type' => 'number'),
                     'ExposureMode' => array('description' => _("Exposure Mode"), 'type' => 'number'),
                     'MeteringMode' => array('description' => _("Metering Mode"), 'type' => 'number'),
                     'Flash' => array('description' => _("Flash Setting"), 'type' => 'number'),
                     'UserComment' => array('description' => _("User Comment"), 'type' => 'text'),
                     'ColorSpace' => array('description' => _("Color Space"), 'type' => 'number'),
                     'SensingMethod' => array('description' => _("Sensing Method"), 'type' => 'number'),
                     'WhiteBalance' => array('description' => _("White Balance"), 'type' => 'number'),
                     'Orientation' => array('description' => _("Camera Orientation"), 'type' => 'number'),
                     'Copyright' => array('description' => _("Copyright"), 'type' => 'text'),
                     'Artist' => array('description' => _("Artist"), 'type' => 'text')
        );
    }

}

/**
 * Library for extracting EXIF information from digital photos.
 *
 * Exifer 1.5
 * Extracts EXIF information from digital photos.
 *
 * Copyright 2005 Jake Olefsky
 * http://www.offsky.com/software/exif/index.php
 *
 * @author Jake Olefsky <jake@olefsky.com>
 * @package Ansel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more
 * details. http://www.gnu.org/copyleft/gpl.html
 *
 * This script will correctly parse all of the EXIF data included in
 * images taken with digital cameras.  It will read the IDF0, IDF1,
 * SubIDF and InteroperabilityIFD fields as well as parsing some of
 * the MakerNote fields that vary depending on camera make and model.
 * This script parses more tags than the internal PHP exif
 * implementation and it will correctly identify and decode what all
 * the values mean.
 *
 * Tested with:
 * Nikon CoolPix 700
 * Nikon CoolPix 4500
 * Nikon CoolPix 950
 * Nikon Coolpix 5700
 * Canon PowerShot S200
 * Canon PowerShot S110
 * Olympus C2040Z
 * Olympus C960
 * Canon Ixus
 * Canon EOS 300D
 * Canon Digital Rebel
 * Canon EOS 10D
 * Canon PowerShot G2
 * FujiFilm DX 10
 * FujiFilm MX-1200
 * FujiFilm FinePix2400
 * FujiFilm FinePix2600
 * FujiFilm FinePix S602
 * FujiFilm FinePix40i
 * Sony D700
 * Sony Cybershot
 * Kodak DC210
 * Kodak DC240
 * Kodak DC4800
 * Kodak DX3215
 * Ricoh RDC-5300
 * Sanyo VPC-G250
 * Sanyo VPC-SX550
 * Epson 3100z
 */

/**
 * Converts from Intel to Motorola endian. Just reverses the bytes
 * (assumes hex is passed in).
 */
function intel2Moto($intel)
{
    $len = strlen($intel);
    $moto = '';
    for ($i = 0; $i <= $len; $i += 2) {
        $moto .= substr($intel, $len - $i, 2);
    }
    return $moto;
}

/**
 * Looks up the name of the tag.
 */
function lookup_tag($tag)
{
    $ExifTagMap = array(
        // IFD0 Camera Tags
        '000b' => 'ACDComment',                     // text string up to 999 bytes long
        '00fe' => 'ImageType',                      // integer -2147483648 to 2147483647
        '0106' => 'PhotometricInterpret',
        '010e' => 'ImageDescription',               // text string up to 999 bytes long
        '010f' => 'Make',                           // text string up to 999 bytes long
        '0110' => 'Model',                          // text string up to 999 bytes long
        '0112' => 'Orientation',                    // integer values 1-9
        '0115' => 'SamplePerPixel',                 // integer 0-65535
        '011a' => 'xResolution',                    // positive rational number
        '011b' => 'yResolution',                    // positive rational number
        '011c' => 'PlanarConfig',                   // integer values 1-2
        '0128' => 'ResolutionUnit',                 // integer values 1-3
        '0131' => 'Software',                       // text string up to 999 bytes long
        '0132' => 'DateTime',                       // YYYY:MM:DD HH:MM:SS
        '013b' => 'Artist',                         // text string up to 999 bytes long
        '013c' => 'HostComputer',                   // text string
        '013e' => 'WhitePoint',                     // two positive rational numbers
        '013f' => 'PrimaryChromaticities',          // six positive rational numbers
        '0211' => 'YCbCrCoefficients',              // three positive rational numbers
        '0213' => 'YCbCrPositioning',               // integer values 1-2
        '0214' => 'ReferenceBlackWhite',            // six positive rational numbers
        '8298' => 'Copyright',                      // text string up to 999 bytes long
        '8649' => 'PhotoshopSettings',
        '8825' => 'GPSInfoOffset',
        '9286' => 'UserCommentOld',
        '8769' => 'ExifOffset',                     // positive integer

        // SubIFD Image Tags
        '829a' => 'ExposureTime',                   // seconds or fraction of seconds 1/x
        '829d' => 'FNumber',                        // positive rational number
        '8822' => 'ExposureProgram',                // integer value 1-9
        '8824' => 'SpectralSensitivity',
        '8827' => 'ISOSpeedRatings',                // integer 0-65535
        '9000' => 'ExifVersion',
        '9003' => 'DateTimeOriginal',               // YYYY:MM:DD HH:MM:SS
        '9004' => 'DateTimedigitized',              // YYYY:MM:DD HH:MM:SS
        '9101' => 'ComponentsConfiguration',
        '9102' => 'CompressedBitsPerPixel',         // positive rational number
        '9201' => 'ShutterSpeedValue',              // seconds or fraction of seconds 1/x
        '9202' => 'ApertureValue',                  // positive rational number
        '9203' => 'BrightnessValue',                // positive rational number
        '9204' => 'ExposureBiasValue',              // positive rational number (EV)
        '9205' => 'MaxApertureValue',               // positive rational number
        '9206' => 'SubjectDistance',                // positive rational number (meters)
        '9207' => 'MeteringMode',                   // integer 1-6 and 255
        '9208' => 'LightSource',                    // integer 1-255
        '9209' => 'Flash',                          // integer 1-255
        '920a' => 'FocalLength',                    // positive rational number (mm)
        '9213' => 'ImageHistory',                   // text string up to 999 bytes long
        '927c' => 'MakerNote',                      // a bunch of data
        '9286' => 'UserComment',                    // text string
        '9290' => 'SubsecTime',                     // text string up to 999 bytes long
        '9291' => 'SubsecTimeOriginal',             // text string up to 999 bytes long
        '9292' => 'SubsecTimeDigitized',            // text string up to 999 bytes long
        'a000' => 'FlashPixVersion',
        'a001' => 'ColorSpace',                     // values 1 or 65535
        'a002' => 'ExifImageWidth',                 // integer 1-65535
        'a003' => 'ExifImageHeight',                // integer 1-65535
        'a004' => 'RelatedSoundFile',               // text string 12 bytes long
        'a005' => 'ExifInteroperabilityOffset',     // positive integer
        'a20c' => 'SpacialFreqResponse',
        'a20b' => 'FlashEnergy',                    // positive rational number
        'a20e' => 'FocalPlaneXResolution',          // positive rational number
        'a20f' => 'FocalPlaneYResolution',          // positive rational number
        'a210' => 'FocalPlaneResolutionUnit',       // values 1-3
        'a214' => 'SubjectLocation',                // two integers 0-65535
        'a215' => 'ExposureIndex',                  // positive rational number
        'a217' => 'SensingMethod',                  // values 1-8
        'a300' => 'FileSource',                     // integer
        'a301' => 'SceneType',                      // integer
        'a302' => 'CFAPattern',                     // undefined data type
        'a401' => 'CustomerRender',                 // values 0 or 1
        'a402' => 'ExposureMode',                   // values 0-2
        'a403' => 'WhiteBalance',                   // values 0 or 1
        'a404' => 'DigitalZoomRatio',               // positive rational number
        'a406' => 'SceneCaptureMode',               // values 0-3
        'a407' => 'GainControl',                    // values 0-4
        'a408' => 'Contrast',                       // values 0-2
        'a409' => 'Saturation',                     // values 0-2
        'a40a' => 'Sharpness',                      // values 0-2

        // Interoperability IFD
        '0001' => 'InteroperabilityIndex',          // text string 3 bytes long
        '0002' => 'InteroperabilityVersion',        // datatype undefined
        '1000' => 'RelatedImageFileFormat',         // text string up to 999 bytes long
        '1001' => 'RelatedImageWidth',              // integer in range 0-65535
        '1002' => 'RelatedImageLength',             // integer in range 0-65535

        // IFD1 Thumbnail
        '0100' => 'ImageWidth',                     // integer in range 0-65535
        '0101' => 'ImageLength',                    // integer in range 0-65535
        '0102' => 'BitsPerSample',                  // integers in range 0-65535
        '0103' => 'Compression',                    // values 1 or 6
        '0106' => 'PhotometricInterpretation',      // values 0-4
        '010e' => 'ThumbnailDescription',           // text string up to 999 bytes long
        '010f' => 'ThumbnailMake',                  // text string up to 999 bytes long
        '0110' => 'ThumbnailModel',                 // text string up to 999 bytes long
        '0111' => 'StripOffsets',
        '0112' => 'ThumbnailOrientation',           // integer 1-9
        '0115' => 'SamplesPerPixel',
        '0116' => 'RowsPerStrip',
        '0117' => 'StripByteCounts',
        '011a' => 'ThumbnailXResolution',           // positive rational number
        '011b' => 'ThumbnailYResolution',           // positive rational number
        '011c' => 'PlanarConfiguration',            // values 1 or 2
        '0128' => 'ThumbnailResolutionUnit',        // values 1-3
        '0201' => 'JpegIFOffset',
        '0202' => 'JpegIFByteCount',
        '0212' => 'YCbCrSubSampling',

        // Misc.
        '00ff' => 'SubfileType',
        '012d' => 'TransferFunction',
        '013d' => 'Predictor',
        '0142' => 'TileWidth',
        '0143' => 'TileLength',
        '0144' => 'TileOffsets',
        '0145' => 'TileByteCounts',
        '014a' => 'SubIFDs',
        '015b' => 'JPEGTables',
        '828d' => 'CFARepeatPatternDim',
        '828e' => 'CFAPattern',
        '828f' => 'BatteryLevel',
        '83bb' => 'IPTC/NAA',
        '8773' => 'InterColorProfile',
        '8828' => 'OECF',
        '8829' => 'Interlace',
        '882a' => 'TimeZoneOffset',
        '882b' => 'SelfTimerMode',
        '920b' => 'FlashEnergy',
        '920c' => 'SpatialFrequencyResponse',
        '920d' => 'Noise',
        '9211' => 'ImageNumber',
        '9212' => 'SecurityClassification',
        '9214' => 'SubjectLocation',
        '9215' => 'ExposureIndex',
        '9216' => 'TIFF/EPStandardID',
        'a20b' => 'FlashEnergy',
    );

    if (isset($ExifTagMap[$tag])) {
        return $ExifTagMap[$tag];
    } else {
        return 'Unknown; ' . $tag;
    }
}

/**
 * Looks up the datatype
 */
function lookup_type(&$type, &$size)
{
    switch ($type) {
    case '0001': $type = 'UBYTE'; $size = 1; break;
    case '0002': $type = 'ASCII'; $size = 1; break;
    case '0003': $type = 'USHORT'; $size = 2; break;
    case '0004': $type = 'ULONG'; $size = 4; break;
    case '0005': $type = 'URATIONAL'; $size = 8; break;
    case '0006': $type = 'SBYTE'; $size = 1; break;
    case '0007': $type = 'UNDEFINED'; $size = 1; break;
    case '0008': $type = 'SSHORT'; $size = 2; break;
    case '0009': $type = 'SLONG'; $size = 4; break;
    case '000a': $type = 'SRATIONAL'; $size = 8; break;
    case '000b': $type = 'FLOAT'; $size = 4; break;
    case '000c': $type = 'DOUBLE'; $size = 8; break;
    default: $type = 'ERROR: ' . $type; $size = 0; break;
    }
    return $type;
}

/**
 * Formats Data for the data type
 */
function formatData($type, $tag, $intel, $data)
{
    if ($type == 'ASCII') {
        // Do nothing.
    } elseif ($type == 'URATIONAL' || $type == 'SRATIONAL') {
        $data = bin2hex($data);
        if ($intel == 1) {
            $data = intel2Moto($data);
        }

        $top = ($intel == 1)
            ? hexdec(substr($data, 8, 8))  // intel stores them bottom-top
            : hexdec(substr($data, 0, 8)); // motorola stores them top-bottom

        $bottom = ($intel == 1)
            ? hexdec(substr($data, 0, 8))  // intel stores them bottom-top
            : hexdec(substr($data, 8, 8)); // motorola stores them top-bottom

        if ($type == 'SRATIONAL' && $top > 2147483647) {
            // This makes the number signed instead of unsigned.
            $top = $top - 4294967296;
        }
        if ($bottom != 0) {
            $data = $top / $bottom;
        } elseif ($top == 0) {
            $data = 0;
        } else {
            $data = $top . '/' . $bottom;
        }

        if ($tag == '9205' || $tag == '9202') {
            // ApertureValue and MaxApertureValue
            // ApertureValue is given in the APEX Mode. Many thanks to Matthieu Froment for this code
            // The formula is : Aperture = 2*log2(FNumber) <=> FNumber = e((Aperture.ln(2))/2)
            // FIXME: need this??
            $data = exp(($data*log(2))/2);
            $data = round($data, 1);//Focal is given with a precision of 1 digit.
        } elseif ($tag=='9201') {
            //ShutterSpeedValue
            // The ShutterSpeedValue is given in the APEX mode. Many thanks to Matthieu Froment for this code
            // The formula is : Shutter = - log2(exposureTime) (Appendix C of EXIF spec.)
            // Where shutter is in APEX, log2(exposure) = ln(exposure)/ln(2)
            // So final formula is : exposure = exp(-ln(2).shutter)
            // The formula can be developed : exposure = 1/(exp(ln(2).shutter))
            // FIXME: need this??
            $data = exp($data * log(2));
            if ($data > 1) $data = floor($data);  //Drop the decimal.
            if ($data > 0) {
                $data = 1/$data; //Final calculation. We now have a floating number. Transform it in a pretty number
                $n=0;$d=0;
                ConvertToFraction($data, $n, $d);
                if ($n>=1 && $d==1) $data = $n; //To avoid exposure times style 3/1 sec.
                else $data = $n.'/'.$d;
            } else {
                $data = 'Bulb';
            }
        }
    } elseif ($type == 'USHORT' || $type=='SSHORT' || $type=='ULONG' || $type=='SLONG' || $type=='FLOAT' || $type=='DOUBLE') {
        $data = bin2hex($data);
        if ($intel == 1) {
            $data = intel2Moto($data);
        } elseif ($intel == 0 && ($type == 'USHORT' || $type == 'SSHORT')) {
            $data = substr($data, 0, 4);
        }
        $data = hexdec($data);

        // Make numbers signed instead of unsigned if necessary.
        if ($type == 'SSHORT' && $data > 32767) {
            $data -= 65536;
        } elseif ($type == 'SLONG' && $data > 2147483647) {
            $data -= 4294967296;
        }

        if ($tag=='0128' || $tag=='a210' || $tag=='0128') {  //ResolutionUnit and FocalPlaneResolutionUnit and ThumbnailResolutionUnit
            if ($data==1) $data = 'No Unit';
            if ($data==2) $data = 'Inch';
            if ($data==3) $data = 'Centimeter';
        } elseif ($tag=='0213') { //YCbCrPositioning
            if ($data==1) $data = 'Center of Pixel Array';
            if ($data==2) $data = 'Datum Point';
        } elseif ($tag=='8822') { //ExposureProgram
            if ($data==1) $data = 'Manual';
            elseif ($data==2) $data = 'Program';
            elseif ($data==3) $data = 'Aperature Priority';
            elseif ($data==4) $data = 'Shutter Priority';
            elseif ($data==5) $data = 'Program Creative';
            elseif ($data==6) $data = 'Program Action';
            elseif ($data==7) $data = 'Portrat';
            elseif ($data==8) $data = 'Landscape';
            else $data = 'Unknown: '.$data;
        } elseif ($tag=='9208') { //LightSource
            if ($data==0) $data = 'Unknown or Auto';
            elseif ($data==1) $data = 'Daylight';
            elseif ($data==2) $data = 'Flourescent';
            elseif ($data==3) $data = 'Tungsten';
            elseif ($data==10) $data = 'Flash';
            elseif ($data==17) $data = 'Standard Light A';
            elseif ($data==18) $data = 'Standard Light B';
            elseif ($data==19) $data = 'Standard Light C';
            elseif ($data==20) $data = 'D55';
            elseif ($data==21) $data = 'D65';
            elseif ($data==22) $data = 'D75';
            elseif ($data==255) $data = 'Other';
            else $data = 'Unknown: '.$data;
        } elseif ($tag=='a001') { //ColorSpace
            if ($data==1) $data = 'sRGB';
            else $data = 'Uncalibrated';
        } elseif ($tag=='a002' || $tag=='a003') { //ExifImageWidth/Height
            $data = $data. ' pixels';
        } elseif ($tag=='0103') { //Compression
            if ($data==1) $data = 'No Compression';
            elseif ($data==6) $data = 'JPEG Compression';
            else $data = 'Unknown: ' . $data;
        } elseif ($tag == '0106') {
            // PhotometricInterpretation
            if ($data == 1) $data = 'Monochrome';
            elseif ($data == 2) $data = 'RGB';
            elseif ($data == 6) $data = 'YCbCr';
            else $data = 'Unknown: ' . $data;
        }
    } elseif ($type == 'UNDEFINED') {
        if ($tag == '9000' || $tag == 'a000' || $tag == '0002') {
            // ExifVersion, FlashPixVersion, InteroperabilityVersion
            $data = 'version ' . $data / 100;
        }
        if ($tag == 'a300') {
            // FileSource
            $data = bin2hex($data);
            $data = str_replace('00', '', $data);
            $data = str_replace('03', 'Digital Still Camera', $data);
        }
        if ($tag == 'a301') {
            // SceneType
            $data = bin2hex($data);
            $data = str_replace('00', '', $data);
            $data = str_replace('01', 'Directly Photographed', $data);
        }
        if ($tag == '9101') {
            // ComponentsConfiguration
            $data = bin2hex($data);
            $data = str_replace(array('01', '02', '03', '04', '05', '06', '00'),
                                array('Y',  'Cb', 'Cr', 'R',  'G',  'B',  ''),
                                $data);
        }
    } else {
        $data = bin2hex($data);
        if ($intel == 1) {
            $data = intel2Moto($data);
        }
    }

    return $data;
}

/**
 * Reads one standard IFD entry.
 */
function read_entry(&$result, $in, $seek, $intel, $ifd_name, $globalOffset)
{
    if (feof($in)) {
        // Test to make sure we can still read.
        $result['Errors']++;
        return;
    }

    // 2 byte tag name.
    $tag = bin2hex(fread($in, 2));
    if ($intel == 1) {
        $tag = intel2Moto($tag);
    }
    $tag_name = lookup_tag($tag);

    // 2 byte datatype.
    $type = bin2hex(fread($in, 2));
    if ($intel == 1) {
        $type = intel2Moto($type);
    }
    lookup_type($type, $size);

    // 4 byte number of elements.
    $count = bin2hex(fread($in, 4));
    if ($intel == 1) {
        $count = intel2Moto($count);
    }
    $bytesofdata = $size * hexdec($count);

    // 4 byte value or pointer to value if larger than 4 bytes.
    $value = fread($in, 4);
    $data = null;

    if ($bytesofdata <= 4) {
        // If datatype is 4 bytes or less, it's the value.
        $data = $value;
    } elseif ($bytesofdata < 100000) {
        // Otherwise it's a pointer to the value, so lets go get it.
        $value = bin2hex($value);
        if ($intel == 1) {
            $value = intel2Moto($value);
        }
        // Offsets are from TIFF header which is 12 bytes from the
        // start of the file.
        $v = fseek($seek, $globalOffset + hexdec($value));
        if ($v == 0) {
            $data = @fread($seek, $bytesofdata);
        } elseif ($v == -1) {
            $result['Errors']++;
        }
    } else {
        // $bytesofdata was too big, so the exif had an error.
        $result['Errors']++;
        return;
    }

    // Remove NUL bytes from data.
    $data = str_replace("\0", '', $data);

    if ($tag_name == 'MakerNote') {
        // If it's a maker tag, we need to parse this specially.
        $make = isset($result['IFD0']['Make']) ? String::lower($result['IFD0']['Make']) : '';

        if (strpos($make, 'nikon') !== false) {
            include_once ANSEL_BASE . '/lib/Exif/nikon.php';
            parseNikon($data, $result);
            $result[$ifd_name]['KnownMaker'] = 1;
        } elseif (strpos($make, 'olympus') !== false) {
            include_once ANSEL_BASE . '/lib/Exif/olympus.php';
            parseOlympus($data, $result, $seek, $globalOffset);
            $result[$ifd_name]['KnownMaker'] = 1;
        } elseif (strpos($make, 'canon') !== false) {
            include_once ANSEL_BASE . '/lib/Exif/canon.php';
            parseCanon($data, $result, $seek, $globalOffset);
            $result[$ifd_name]['KnownMaker'] = 1;
        } elseif (strpos($make, 'fujifilm') !== false) {
            include_once ANSEL_BASE . '/lib/Exif/fujifilm.php';
            parseFujifilm($data, $result);
            $result[$ifd_name]['KnownMaker'] = 1;
        } elseif (strpos($make, 'sanyo') !== false) {
            include_once ANSEL_BASE . '/lib/Exif/sanyo.php';
            parseSanyo($data, $result, $seek, $globalOffset);
            $result[$ifd_name]['KnownMaker'] = 1;
        } else {
            $result[$ifd_name]['KnownMaker'] = 0;
        }
    } elseif ($tag_name == 'GPSInfoOffset') {
        $formatted_data = formatData($type, $tag, $intel, $data);
        $result[$ifd_name]['GPSInfo'] = $formatted_data;
        include_once ANSEL_BASE . '/lib/Exif/gps.php';
        parseGPS($data, $result, $formatted_data, $seek, $globalOffset);
    } else {
        // Format the data depending on the type and tag.
        $result[$ifd_name][$tag_name] = formatData($type, $tag, $intel, $data);
    }
}

/**
 * Pass in a file and this reads the EXIF data.
 *
 * Useful resources
 * http://www.ba.wakwak.com/~tsuruzoh/Computer/Digicams/exif-e.html
 * http://www.w3.org/Graphics/JPEG/jfif.txt
 * http://exif.org/
 * http://www.ozhiker.com/electronics/pjmt/library/list_contents.php4
 * http://www.ozhiker.com/electronics/pjmt/jpeg_info/makernotes.html
 * http://pel.sourceforge.net/
 * http://us2.php.net/manual/en/function.exif-read-data.php
 */
function read_exif_data_raw($path)
{
    if ($path == '' || $path == 'none') {
        return;
    }

    $in = fopen($path, 'rb');
    $seek = fopen($path, 'rb'); // There may be an elegant way to do this with one file handle.

    $globalOffset = 0;

    $result['Errors'] = 0;
    $result['IFD0'] = array();
    $result['SubIFD'] = array();

    if (!$in || !$seek) {
        // If the path was invalid, this error will catch it.
        $result['Errors'] = 1;
        $result['Error'][$result['Errors']] = 'File not found';
        return $result;
    }

    // First 2 bytes of JPEG are 0xFFD8.
    $data = bin2hex(fread($in, 2));
    if ($data == 'ffd8') {
        $result['ValidJpeg'] = 1;
    } else {
        $result['ValidJpeg'] = 0;
        fclose($in);
        fclose($seek);
        return $result;
    }

    $result['ValidIPTCData'] = 0;
    $result['ValidJFIFData'] = 0;
    $result['ValidEXIFData'] = 0;
    $result['ValidAPP2Data'] = 0;
    $result['ValidCOMData'] = 0;

    // Next 2 bytes are MARKER tag (0xFFE#).
    $data = bin2hex(fread($in, 2));
    $size = bin2hex(fread($in, 2));

    // Loop through markers until we get to FFE1 (Exif marker).
    while (!feof($in) && $data != 'ffe1' && $data != 'ffc0' && $data != 'ffd9') {
        if ($data == 'ffe0') {
            // JFIF Marker.
            $result['ValidJFIFData'] = 1;
            $result['JFIF']['Size'] = hexdec($size);

            if (hexdec($size) - 2 > 0) {
                $data = fread($in, hexdec($size) - 2);
                $result['JFIF']['Data'] = $data;
            }

            $result['JFIF']['Identifier'] = substr($data, 0, 5);
            $result['JFIF']['ExtensionCode'] = bin2hex(substr($data, 6, 1));

            $globalOffset += hexdec($size) + 2;
        } elseif ($data == 'ffed') {
            // IPTC Marker.
            $result['ValidIPTCData'] = 1;
            $result['IPTC']['Size'] = hexdec($size);

            if (hexdec($size) - 2 > 0) {
                $data = fread($in, hexdec($size) - 2);
                $result['IPTC']['Data'] = $data;
            }
            $globalOffset += hexdec($size) + 2;
        } elseif ($data == 'ffe2') {
            // EXIF extension Marker.
            $result['ValidAPP2Data'] = 1;
            $result['APP2']['Size'] = hexdec($size);

            if (hexdec($size) - 2 > 0) {
                $data = fread($in, hexdec($size) - 2);
                $result['APP2']['Data'] = $data;
            }
            $globalOffset += hexdec($size) + 2;
        } elseif ($data == 'fffe') {
            // COM extension Marker.
            $result['ValidCOMData'] = 1;
            $result['COM']['Size'] = hexdec($size);

            if (hexdec($size) - 2 > 0) {
                $data = fread($in, hexdec($size) - 2);
                $result['COM']['Data'] = $data;
            }
            $globalOffset += hexdec($size) + 2;
        } elseif ($data == 'ffe1') {
            $result['ValidEXIFData'] = 1;
        }

        $data = bin2hex(fread($in, 2));
        $size = bin2hex(fread($in, 2));
    }

    if ($data == 'ffe1') {
        $result['ValidEXIFData'] = 1;
    } else {
        $result['ValidEXIFData'] = 0;
        fclose($in);
        fclose($seek);
        return $result;
    }

    // Size of APP1.
    $result['APP1Size'] = hexdec($size);

    // Start of APP1 block starts with "Exif" header (6 bytes).
    $header = fread($in, 6);

    // Then there's a TIFF header with 2 bytes of endianess (II or
    // MM).
    $header = fread($in, 2);
    if ($header === 'II') {
        $intel = 1;
        $result['Endian'] = 'Intel';
    } elseif ($header === 'MM') {
        $intel = 0;
        $result['Endian'] = 'Motorola';
    } else {
        $intel = 1;
        // Not sure what the default should be, but this seems
        // reasonable.
        $result['Endian'] = 'Unknown';
    }

    // 2 bytes of 0x002a.
    $tag = bin2hex(fread($in, 2));

    // Then 4 bytes of offset to IFD0 (usually 8 which includes all 8
    // bytes of TIFF header).
    $offset = bin2hex(fread($in, 4));
    if ($intel == 1) {
        $offset = intel2Moto($offset);
    }

    // Check for extremely large values here.
    if (hexdec($offset) > 100000) {
        $result['ValidEXIFData'] = 0;
        fclose($in);
        fclose($seek);
        return $result;
    }

    if (hexdec($offset) > 8) {
        fread($in, hexdec($offset) - 8);
    }

    // Add 12 to the offset to account for TIFF header.
    $globalOffset += 12;

    //===========================================================Start of IFD0
    $num = bin2hex(fread($in, 2));
    if ($intel == 1) {
        $num = intel2Moto($num);
    }
    $num = hexdec($num);
    $result['IFD0NumTags'] = $num;

    if ($num < 1000) {
        // 1000 entries is too much and is probably an error.
        for ($i = 0; $i < $num; $i++) {
            read_entry($result, $in, $seek, $intel, 'IFD0', $globalOffset);
        }
    } else {
        $result['Errors']++;
        $result['Error'][$result['Errors']] = 'Illegal size for IFD0';
    }

    // Store offset to IFD1
    $offset = bin2hex(fread($in, 4));
    if ($intel == 1) $offset = intel2Moto($offset);
    $result['IFD1Offset'] = hexdec($offset);

    // Check for SubIFD
    if (!isset($result['IFD0']['ExifOffset'])
        || $result['IFD0']['ExifOffset'] == 0) {
        fclose($in);
        fclose($seek);
        return $result;
    }

    // Seek to SubIFD (Value of ExifOffset tag) above.
    $ExitOffset = $result['IFD0']['ExifOffset'];
    $v = fseek($in, $globalOffset + $ExitOffset);
    if ($v == -1) {
        $result['Errors']++;
        $result['Error'][$result['Errors']] = 'Couldn\'t Find SubIFD';
    }

    //===========================================================Start of SubIFD
    $num = bin2hex(fread($in, 2));
    if ($intel == 1) {
        $num = intel2Moto($num);
    }
    $num = hexdec($num);
    $result['SubIFDNumTags'] = $num;

    if ($num < 1000) {
        // 1000 entries is too much and is probably an error.
        for ($i = 0; $i < hexdec($num); $i++) {
            read_entry($result, $in, $seek, $intel, 'SubIFD', $globalOffset);
        }
    } else {
        $result['Errors']++;
        $result['Error'][$result['Errors']] = 'Illegal size for SubIFD';
    }

    // Check for IFD1
    if (!isset($result['IFD1Offset']) || $result['IFD1Offset'] == 0) {
        fclose($in);
        fclose($seek);
        return $result;
    }

    // Seek to IFD1
    $v = fseek($in, $globalOffset + $result['IFD1Offset']);
    if ($v == -1) {
        $result['Errors']++;
        $result['Error'][$result['Errors']] = 'Couldn\'t Find IFD1';
    }

    //===========================================================Start of IFD1
    $num = bin2hex(fread($in, 2));
    if ($intel == 1) {
        $num = intel2Moto($num);
    }
    $num = hexdec($num);
    $result['IFD1NumTags'] = $num;

    if ($num < 1000) {
        // 1000 entries is too much and is probably an error.
        for ($i = 0; $i < hexdec($num); $i++) {
            read_entry($result, $in, $seek, $intel, 'IFD1', $globalOffset);
        }
    } else {
        $result['Errors']++;
        $result['Error'][$result['Errors']] = 'Illegal size for IFD1';
    }

    // If we can, include the raw thumbnail data.
    if ($result['IFD1']['JpegIFOffset'] > 0 && $result['IFD1']['JpegIFByteCount'] > 0) {
        $data = null;
        $v = fseek($seek, $globalOffset + $result['IFD1']['JpegIFOffset']);
        if ($v == 0) {
            $data = fread($seek, $result['IFD1']['JpegIFByteCount']);
        } elseif ($v == -1) {
            $result['Errors']++;
        }
        $result['IFD1']['ThumbnailData'] = $data;
    }

    // Check for Interoperability IFD
    if (!isset($result['SubIFD']['ExifInteroperabilityOffset'])
        || $result['SubIFD']['ExifInteroperabilityOffset'] == 0) {
        fclose($in);
        fclose($seek);
        return $result;
    }

    // Seek to InteroperabilityIFD
    $v = fseek($in, $globalOffset + $result['SubIFD']['ExifInteroperabilityOffset']);
    if ($v == -1) {
        $result['Errors']++;
        $result['Error'][$result['Errors']] = 'Couldn\'t Find InteroperabilityIFD';
    }

    //===========================================================Start of InteroperabilityIFD
    $num = bin2hex(fread($in, 2));
    if ($intel == 1) {
        $num = intel2Moto($num);
    }
    $num = hexdec($num);
    $result['InteroperabilityIFDNumTags'] = $num;

    if ($num < 1000) {
        // 1000 entries is too much and is probably an error.
        for ($i = 0; $i < hexdec($num); $i++) {
            read_entry($result, $in, $seek, $intel, 'InteroperabilityIFD', $globalOffset);
        }
    } else {
        $result['Errors']++;
        $result['Error'][$result['Errors']] = 'Illegal size for InteroperabilityIFD';
    }

    fclose($in);
    fclose($seek);
    return $result;
}

/**
 * Converts a floating point number into a fraction.  Many thanks to Matthieu
 * Froment for this code
 */
function ConvertToFraction($v, &$n, &$d)
{
    $MaxTerms = 15;         // Limit to prevent infinite loop
    $MinDivisor = 0.000001; // Limit to prevent divide by zero
    $MaxError = 0.00000001; // How close is enough

    $f = $v; // Initialize fraction being converted

    $n_un = 1; // Initialize fractions with 1/0, 0/1
    $d_un = 0;
    $n_deux = 0;
    $d_deux = 1;

    for ($i = 0; $i < $MaxTerms; $i++) {
        $a = floor($f); // Get next term
        $f = $f - $a; // Get new divisor
        $n = $n_un * $a + $n_deux; // Calculate new fraction
        $d = $d_un * $a + $d_deux;
        $n_deux = $n_un; // Save last two fractions
        $d_deux = $d_un;
        $n_un = $n;
        $d_un = $d;

        if ($f < $MinDivisor) {
            // Quit if dividing by zero.
            break;
        }

        if (abs($v - $n / $d) < $MaxError) {
            break;
        }

        // Take reciprocal.
        $f = 1 / $f;
    }
}
